import logging
import anyio

from distmqtt.client import open_mqttclient
from distmqtt.mqtt.constants import QOS_1, QOS_2


#
# This sample shows how to publish messages to broker using different QOS
# Debug outputs shows the message flows
#

logger = logging.getLogger(__name__)

config = {
    'will': {
        'topic': '/will/client',
        'message': b'Dead or alive',
        'qos': 0x01,
        'retain': True
    },
    'capath': '.',
}


async def test_coro():
    async with open_mqttclient(config=config) as C:
        await C.connect('wss://test.mosquitto.org:8081/', cafile='mosquitto.org.crt')
        async with anyio.create_task_group() as tg:
            await tg.spawn(C.publish('a/b', b'TEST MESSAGE WITH QOS_0')
            await tg.spawn(C.publish('a/b', b'TEST MESSAGE WITH QOS_1', qos=QOS_1)
            await tg.spawn(C.publish('a/b', b'TEST MESSAGE WITH QOS_2', qos=QOS_2)
        logger.info("messages published")


if __name__ == '__main__':
    formatter = "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.DEBUG, format=formatter)
    anyio.run(test_coro)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
from distmqtt.errors import DistMQTTException
from distmqtt.mqtt.packet import (
    CONNECT, CONNACK, PUBLISH, PUBACK, PUBREC, PUBREL, PUBCOMP, SUBSCRIBE,
    SUBACK, UNSUBSCRIBE, UNSUBACK, PINGREQ, PINGRESP, DISCONNECT,
    MQTTFixedHeader)
from distmqtt.mqtt.connect import ConnectPacket
from distmqtt.mqtt.connack import ConnackPacket
from distmqtt.mqtt.disconnect import DisconnectPacket
from distmqtt.mqtt.pingreq import PingReqPacket
from distmqtt.mqtt.pingresp import PingRespPacket
from distmqtt.mqtt.publish import PublishPacket
from distmqtt.mqtt.puback import PubackPacket
from distmqtt.mqtt.pubrec import PubrecPacket
from distmqtt.mqtt.pubrel import PubrelPacket
from distmqtt.mqtt.pubcomp import PubcompPacket
from distmqtt.mqtt.subscribe import SubscribePacket
from distmqtt.mqtt.suback import SubackPacket
from distmqtt.mqtt.unsubscribe import UnsubscribePacket
from distmqtt.mqtt.unsuback import UnsubackPacket

packet_dict = {
    CONNECT: ConnectPacket,
    CONNACK: ConnackPacket,
    PUBLISH: PublishPacket,
    PUBACK: PubackPacket,
    PUBREC: PubrecPacket,
    PUBREL: PubrelPacket,
    PUBCOMP: PubcompPacket,
    SUBSCRIBE: SubscribePacket,
    SUBACK: SubackPacket,
    UNSUBSCRIBE: UnsubscribePacket,
    UNSUBACK: UnsubackPacket,
    PINGREQ: PingReqPacket,
    PINGRESP: PingRespPacket,
    DISCONNECT: DisconnectPacket
}


def packet_class(fixed_header: MQTTFixedHeader):
    try:
        cls = packet_dict[fixed_header.packet_type]
        return cls
    except KeyError:
        raise DistMQTTException("Unexpected packet Type '%s'" % fixed_header.packet_type)



import logging
import anyio
import os
from distmqtt.broker import create_broker

logger = logging.getLogger(__name__)

config = {
    'listeners': {
        'default': {
            'type': 'tcp',
            'bind': '0.0.0.0:1883',
        },
        'ws-mqtt': {
            'bind': '127.0.0.1:8080',
            'type': 'ws',
            'max_connections': 10,
        },
    },
    'sys_interval': 10,
    'auth': {
        'allow-anonymous': True,
        'password-file': os.path.join(os.path.dirname(os.path.realpath(__file__)), "passwd"),
        'plugins': [
            'auth_file', 'auth_anonymous'
        ]

    },
    'topic-check': {
        'enabled': True,
        'plugins': [
            'topic_taboo'
        ]
    }
}

async def test_coro():
    async with create_broker(config) as broker:
        while True:
            await anyio.sleep(99999)


if __name__ == '__main__':
    formatter = "[%(asctime)s] :: %(levelname)s :: %(name)s :: %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)
    anyio.run(test_coro)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
"""
distmqtt_pub - MQTT 3.1.1 publisher

Usage:
    distmqtt_pub --version
    distmqtt_pub (-h | --help)
    distmqtt_pub --url BROKER_URL -t TOPIC (-f FILE | -l | -m MESSAGE | -M message | -n | -s | -S) [-C codec] [-c CONFIG_FILE] [-i CLIENT_ID] [-q | --qos QOS] [-d] [-k KEEP_ALIVE] [--clean-session] [--ca-file CAFILE] [--ca-path CAPATH] [--ca-data CADATA] [ --will-topic WILL_TOPIC [--will-message WILL_MESSAGE] [--will-qos WILL_QOS] [--will-retain] ] [--extra-headers HEADER] [-r]

Options:
    -h --help           Show this screen.
    --version           Show version.
    --url BROKER_URL    Broker connection URL (musr conform to MQTT URI scheme (see https://github.com/mqtt/mqtt.github.io/wiki/URI-Scheme>)
    -c CONFIG_FILE      Broker configuration file (YAML format)
    -i CLIENT_ID        Id to use as client ID.
    -q | --qos QOS      Quality of service to use for the message, from 0, 1 and 2. Defaults to 0.
    -r                  Set retain flag on connect
    -t TOPIC            Message topic
    -m MESSAGE          Message data to send
    -M MESSAGE          data to evaluate and send
    -f FILE             Read file by line and publish message for each line
    -s                  Read from stdin and publish message for each line
    -S                  Read from stdin, eval, send the result
    -C codec            Codec to send the result with
    -k KEEP_ALIVE       Keep alive timeout in second
    --clean-session     Clean session on connect (defaults to False)
    --ca-file CAFILE]   CA file
    --ca-path CAPATH]   CA Path
    --ca-data CADATA    CA data
    --will-topic WILL_TOPIC
    --will-message WILL_MESSAGE
    --will-qos WILL_QOS
    --will-retain
    --extra-headers EXTRA_HEADERS      JSON object with key-value pairs of additional headers for websocket connections
    -d                  Enable debug messages
"""

import sys
import logging
import anyio
import os
import json
from distmqtt.client import open_mqttclient, ConnectException, _codecs
from distmqtt.version import get_version
from docopt import docopt
from distmqtt.utils import read_yaml_config


logger = logging.getLogger(__name__)


def _gen_client_id():
    import os
    import socket
    pid = os.getpid()
    hostname = socket.gethostname()
    return "distmqtt_pub/%d-%s" % (pid, hostname)


def _get_qos(arguments):
    try:
        return int(arguments['--qos'][0])
    except:
        return None

def _get_extra_headers(arguments):
    try:
        return json.loads(arguments['--extra-headers'])
    except:
        return {}

def _get_message(arguments):
    codec = arguments['-C'] or "utf-8"
    codec = _codecs[codec]()

    if arguments['-n']:
        yield b''
    if arguments['-m']:
        yield arguments['-m'].encode(encoding='utf-8')
    if arguments['-M']:
        yield codec.encode(eval(arguments['-M']))
    if arguments['-f']:
        try:
            with open(arguments['-f'], 'r') as f:
                for line in f:
                    yield line.encode(encoding='utf-8')
        except:
            logger.error("Failed to read file '%s'", arguments['-f'])
    if arguments['-l']:
        import sys
        for line in sys.stdin:
            if line:
                yield line.encode(encoding='utf-8')
    if arguments['-s']:
        import sys
        message = bytearray()
        for line in sys.stdin:
            message.extend(line.encode(encoding='utf-8'))
        yield message
    if arguments['-S']:
        import sys
        message = sys.stdin.read()
        yield codecs.encode(eval(message))


async def do_pub(client, arguments):
    logger.info("%s Connecting to broker", client.client_id)

    await client.connect(uri=arguments['--url'],
                                cleansession=arguments['--clean-session'],
                                cafile=arguments['--ca-file'],
                                capath=arguments['--ca-path'],
                                cadata=arguments['--ca-data'],
                                extra_headers=_get_extra_headers(arguments))
    try:
        qos = _get_qos(arguments)
        topic = arguments['-t']
        retain = arguments['-r']
        async with anyio.create_task_group() as tg:
            for message in _get_message(arguments):
                logger.info("%s Publishing to '%s'", client.client_id, topic)
                await tg.spawn(client.publish, topic, message, qos, retain)
        logger.info("%s Disconnected from broker", client.client_id)
    except KeyboardInterrupt:
        logger.info("%s Disconnected from broker", client.client_id)
    except ConnectException as ce:
        logger.fatal("connection to '%s' failed: %r", arguments['--url'], ce)
    finally:
        await client.disconnect()


async def main(*args, **kwargs):
    arguments = docopt(__doc__, version=get_version())
    #print(arguments)
    formatter = "[%(asctime)s] :: %(levelname)s - %(message)s"

    if arguments['-d']:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(level=level, format=formatter)

    config = None
    if arguments['-c']:
        config = read_yaml_config(arguments['-c'])
    else:
        config = read_yaml_config(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'default_client.yaml'))
        logger.debug("Using default configuration")

    client_id = arguments.get("-i", None)
    if not client_id:
        client_id = _gen_client_id()

    if arguments['-k']:
        config['keep_alive'] = int(arguments['-k'])

    if arguments['--will-topic'] and arguments['--will-message'] and arguments['--will-qos']:
        config['will'] = dict()
        config['will']['topic'] = arguments['--will-topic']
        config['will']['message'] = arguments['--will-message'].encode('utf-8')
        config['will']['qos'] = int(arguments['--will-qos'])
        config['will']['retain'] = arguments['--will-retain']

    async with open_mqttclient(client_id=client_id, config=config) as C:
        await do_pub(C, arguments)


if __name__ == "__main__":
    anyio.run(main)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import anyio
import unittest

from distmqtt.mqtt.pubrel import PubrelPacket, PacketIdVariableHeader
from distmqtt.adapters import BufferAdapter


class PubrelPacketTest(unittest.TestCase):
    def test_from_stream(self):
        data = b'\x60\x02\x00\x0a'
        stream = BufferAdapter(data)
        message = anyio.run(PubrelPacket.from_stream, stream)
        self.assertEqual(message.variable_header.packet_id, 10)

    def test_to_bytes(self):
        variable_header = PacketIdVariableHeader(10)
        publish = PubrelPacket(variable_header=variable_header)
        out = publish.to_bytes()
        self.assertEqual(out, b'\x62\x02\x00\x0a')



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
from distmqtt.mqtt.packet import MQTTPacket, MQTTFixedHeader, PUBACK, PacketIdVariableHeader
from distmqtt.errors import DistMQTTException


class PubackPacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = None

    @property
    def packet_id(self):
        return self.variable_header.packet_id

    @packet_id.setter
    def packet_id(self, val: int):
        self.variable_header.packet_id = val

    def __init__(self, fixed: MQTTFixedHeader=None, variable_header: PacketIdVariableHeader=None):
        if fixed is None:
            header = MQTTFixedHeader(PUBACK, 0x00)
        else:
            if fixed.packet_type is not PUBACK:
                raise DistMQTTException("Invalid fixed packet type %s for PubackPacket init" % fixed.packet_type)
            header = fixed
        super().__init__(header)
        self.variable_header = variable_header
        self.payload = None

    @classmethod
    def build(cls, packet_id: int):
        v_header = PacketIdVariableHeader(packet_id)
        packet = PubackPacket(variable_header=v_header)
        return packet



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import anyio
import logging

from typing import Optional

from transitions import Machine
from collections import OrderedDict

from distmqtt.mqtt.publish import PublishPacket
from distmqtt.errors import DistMQTTException, MQTTException
from distmqtt.mqtt.constants import QOS_0

from distmqtt.broker import Broker
from distmqtt.session import ApplicationMessage

from distkv.client import open_client as open_distkv_client
from distkv.util import PathLongener


class DistKVbroker(Broker):
    """
    A Broker that routes messages through DistKV / MQTT.
    """

    __slots__ = (
        '_distkv_broker__client',
        '_distkv_broker__topic',
    )
    __retain = False

    def __init__(self, tg: anyio.abc.TaskGroup, config=None, plugin_namespace=None):
        self.__client = None
        super().__init__(tg, config=config, plugin_namespace=plugin_namespace)

    async def __session(self, cfg: dict, evt: Optional[anyio.abc.Event] = None):
        async with open_distkv_client(**cfg['server']) as client:
            self.__client = client
            self.__topic = cfg['topic']

            try:
                async with self.__client.serf_mon(tag=self.__topic) as q:
                    await evt.set()
                    async for m in q:
                        d = m.data
                        sess = d.pop('session', None)
                        if sess is not None:
                            sess = self._sessions.get(sess, None)
                            if sess is not None:
                                sess = sess[0]
                        await super().broadcast_message(session=sess, **d)
            finally:
                self.__client = None

    async def __retain(self, cfg: dict, evt: Optional[anyio.abc.Event] = None):
        path = cfg['retain']
        if isinstance(path,str):
            path = path.split('/')
        self.__retain = list(path)
        pl = PathLongener(())  # intentionally not including the path
        async with self.__client.watch(*path, fetch=True, long_path=False) as w:
            await evt.set()
            async for msg in w:
                if 'path' not in msg or msg.get('value', None) is None:
                    continue
                pl(msg)
                topic = '/'.join(msg.path)
                data = msg.get('value', None)
                super().retain_message(None, topic, data, qos=None)
                if not data:
                    continue
                await super().broadcast_message(session=None, topic=topic, data=data)
            
    async def start(self):
        cfg = self.config['distkv']

        evt = anyio.create_event()
        await self._tg.spawn(self.__session, cfg, evt)
        await evt.wait()
        if 'retain' in cfg:
            evt = anyio.create_event()
            await self._tg.spawn(self.__retain, cfg, evt)
            await evt.wait()
        await super().start()

    async def broadcast_message(self, session, topic, data, force_qos=None, qos=None, retain=False):
        if retain and self.__retain:
            await self.__client.set(*(self.__retain + topic.split('/')), value=data)
            return
        msg = dict(session=session.client_id, topic=topic, data=data, qos=qos, force_qos=force_qos, retain=retain)
        await self.__client.serf_send(tag=self.__topic, data=msg)




# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.

import anyio
import logging
import ssl
import copy
from urllib.parse import urlparse, urlunparse
from functools import wraps
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager
from wsproto.utilities import ProtocolError
from asyncwebsockets import create_websocket

from distmqtt.utils import not_in_dict_or_none, match_topic
from distmqtt.session import Session
from distmqtt.errors import NoDataException
from distmqtt.mqtt.connack import CONNECTION_ACCEPTED
from distmqtt.mqtt.protocol.client_handler import ClientProtocolHandler
from distmqtt.adapters import StreamAdapter, WebSocketsAdapter
from distmqtt.plugins.manager import PluginManager, BaseContext
from distmqtt.mqtt.protocol.handler import ProtocolHandlerException
from distmqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from distmqtt import codecs
from collections import deque


_defaults = {
    'keep_alive': 10,
    'ping_delay': 1,
    'default_qos': 0,
    'default_retain': False,
    'auto_reconnect': True,
    'reconnect_max_interval': 10,
    'reconnect_retries': 2,
    'codec': 'noop',
}

_codecs = {
    'noop': codecs.NoopCodec,
    'no-op': codecs.NoopCodec,
    'utf8': codecs.UTF8Codec,
    'utf-8': codecs.UTF8Codec,
    'msgpack': codecs.MsgPackCodec,
}

_handler_id = 0

class _set(set):
    pass

class ClientException(Exception):
    pass


class ConnectException(ClientException):
    pass


class ClientContext(BaseContext):
    """
        ClientContext is used as the context passed to plugins interacting with the client.
        It act as an adapter to client services from plugins
    """
    def __init__(self):
        super().__init__()
        self.config = None


base_logger = logging.getLogger(__name__)


def mqtt_connected(func):
    """
        MQTTClient coroutines decorator which will wait until connection before calling the decorated method.
        :param func: coroutine to be called once connected
        :return: coroutine result
    """
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        if not self._connected_state.is_set():
            base_logger.warning("Client not connected, waiting for it")
            async with anyio.create_task_group() as tg:
                async def wait_connected():
                    await self._connected_state.wait()
                    await tg.cancel_scope.cancel()
                async def wait_no_more():
                    await self._no_more_connections.wait()
                    raise ClientException("Will not reconnect")
                await tg.spawn(wait_connected)
                await tg.spawn(wait_no_more)
        return await func(self, *args, **kwargs)
    return wrapper

@asynccontextmanager
async def open_mqttclient(client_id=None, config=None, codec=None):
    """
        MQTT client implementation.

        MQTTClient instances provides API for connecting to a broker and send/receive messages using the MQTT protocol.

        :param client_id: MQTT client ID to use when connecting to the broker. If none, it will generated randomly by :func:`distmqtt.utils.gen_client_id`
        :param config: Client configuration
        :param codec: Codec to default to, the config or "no-op" if not given.
        :return: async context manager returning a class instance

        Example usage::

            async with open_mqttclient(config=dict(uri="mqtt://my-broker.example")) as client:
                # await client.connect("mqtt://my-broker.example")  # alternate use
                await C.subscribe([
                        ('$SYS/broker/uptime', QOS_1),
                        ('$SYS/broker/load/#', QOS_2),
                    ])
                async for msg in client:
                    packet = message.publish_packet
                    print("%d:  %s => %s" % (i, packet.variable_header.topic_name, str(packet.payload.data)))

    """
    async with anyio.create_task_group() as tg:
        C = MQTTClient(tg, client_id, config=config, codec=codec)
        try:
            if 'uri' in config:
                await C.connect(**config)
            yield C
        finally:
            async with anyio.open_cancel_scope(shield=True):
                await C.disconnect()
                await tg.cancel_scope.cancel()


class MQTTClient:
    """
        MQTT client implementation.

        MQTTClient instances provides API for connecting to a broker and send/receive messages using the MQTT protocol.

        :param tg: The task group in which to run open-ended subtasks.
        :param client_id: MQTT client ID to use when connecting to the broker. If none, it will generated randomly by :func:`distmqtt.utils.gen_client_id`
        :param config: Client configuration
        :param codec: Codec to default to, the config or "no-op" if not given.
        :return: class instance

        You should use :func:`open_mqttclient` to create an instance of
        this class.
    """

    def __init__(self, tg: anyio.abc.TaskGroup, client_id=None, config=None, codec=None):
        self.logger = logging.getLogger(__name__)
        self.config = copy.deepcopy(_defaults)
        if config is not None:
            self.config.update(config)
        if client_id is not None:
            self.client_id = client_id
        else:
            from distmqtt.utils import gen_client_id
            self.client_id = gen_client_id()
            self.logger.debug("Using generated client ID : %s", self.client_id)

        self.session = None
        self._tg = tg
        self._handler = None
        self._disconnect_task = None
        self._connected_state = anyio.create_event()
        self._no_more_connections = anyio.create_event()
        self.extra_headers = {}
        self.codec = _codecs[codec or self.config['codec']]()

        self._subscriptions = None

        # Init plugins manager
        context = ClientContext()
        context.config = self.config
        self.plugins_manager = PluginManager(tg, 'distmqtt.client.plugins', context)
        self.client_task = None

    async def connect(self,
                uri=None,
                cleansession=None,
                cafile=None,
                capath=None,
                cadata=None,
                extra_headers={},
                **kw):
        """
            Connect to a remote broker.

            At first, a network connection is established with the server using the given protocol (``mqtt``, ``mqtts``, ``ws`` or ``wss``). Once the socket is connected, a `CONNECT <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718028>`_ message is sent with the requested informations.

            This method is a *coroutine*.

            :param uri: Broker URI connection, conforming to `MQTT URI scheme <https://github.com/mqtt/mqtt.github.io/wiki/URI-Scheme>`_. Uses ``uri`` config attribute by default.
            :param cleansession: MQTT CONNECT clean session flag
            :param cafile: server certificate authority file (optional, used for secured connection)
            :param capath: server certificate authority path (optional, used for secured connection)
            :param cadata: server certificate authority data (optional, used for secured connection)
            :param extra_headers: a dictionary with additional http headers that should be sent on the initial connection (optional, used only with websocket connections)
            :return: `CONNACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718033>`_ return code
            :raise: :class:`distmqtt.client.ConnectException` if connection fails
        """
        self.session = self._initsession(uri, cleansession, cafile, capath, cadata)
        self.extra_headers = extra_headers;
        self.logger.debug("Connect to: %s", uri)

        try:
            return await self._do_connect()
        except SyntaxError as be:
            self.logger.warning("Connection failed: %r", be)
            auto_reconnect = self.config.get('auto_reconnect', False)
            if not auto_reconnect:
                raise
            else:
                return await self.reconnect()

    async def disconnect(self):
        """
            Disconnect from the connected broker.

            This method sends a `DISCONNECT <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718090>`_ message and closes the network socket.

            This method is a *coroutine*.
        """

        # do not reconnect any more
        self.config['auto_reconnect'] = False
        await self.cancel_tasks()

        if self.session is not None and self.session.transitions.is_connected():
            if self._disconnect_task is not None:
                await self._disconnect_task.cancel()
            await self._handler.mqtt_disconnect()
            self._connected_state.clear()
            await self._handler.stop()
            self.session.transitions.disconnect()
        else:
            self.logger.warning("Client session is not currently connected, ignoring call")

    async def cancel_tasks(self):
        """
        Before disconnection need to cancel all pending tasks
        :return:
        """
        if self.client_task is not None:
            task,self.client_task = self.client_task, None
            await task.cancel()

    async def reconnect(self, cleansession=None):
        """
            Reconnect a previously connected broker.

            Reconnection tries to establish a network connection and send a `CONNECT <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718028>`_ message.
            Retries interval and attempts can be controled with the ``reconnect_max_interval`` and ``reconnect_retries`` configuration parameters.

            This method is a *coroutine*.

            :param cleansession: clean session flag used in MQTT CONNECT messages sent for reconnections.
            :return: `CONNACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718033>`_ return code
            :raise: :class:`distmqtt.client.ConnectException` if re-connection fails after max retries.
        """

        if self.session.transitions.is_connected():
            self.logger.warning("Client already connected")
            return CONNECTION_ACCEPTED

        if cleansession:
            self.session.clean_session = cleansession
        self.logger.debug("Reconnecting with session parameters: %r", self.session)
        reconnect_max_interval = self.config.get('reconnect_max_interval', 10)
        reconnect_retries = self.config.get('reconnect_retries', 5)
        nb_attempt = 1
        await anyio.sleep(1)
        while True:
            try:
                self.logger.debug("Reconnect attempt %d ...", nb_attempt)
                return await self._do_connect()
            except SyntaxError as e:  # no you can't reconnect on every exception
                self.logger.warning("Reconnection attempt failed: %r", e)
                if reconnect_retries >= 0 and nb_attempt > reconnect_retries:
                    self.logger.error("Maximum number of connection attempts reached. Reconnection aborted")
                    raise ConnectException("Too many connection attempts failed")
                exp = 2 ** nb_attempt
                delay = exp if exp < reconnect_max_interval else reconnect_max_interval
                self.logger.debug("Waiting %d second before next attempt", delay)
                await anyio.sleep(delay)
                nb_attempt += 1

    async def _do_connect(self):
        return_code = await self._connect_coro()
        evt = anyio.create_event()
        await self._tg.spawn(self.handle_connection_close, evt)
        await evt.wait()
        return return_code

    @mqtt_connected
    async def ping(self):
        """
            Ping the broker.

            Send a MQTT `PINGREQ <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718081>`_ message for response.

            This method is a *coroutine*.
        """

        if self.session.transitions.is_connected():
            await self._handler.mqtt_ping()
        else:
            self.logger.warning("MQTT PING request incompatible with current session state '%s'",
                                self.session.transitions.state)

    @mqtt_connected
    async def publish(self, topic, message, qos=None, retain=None, codec=None):
        """
            Publish a message to the broker.

            Send a MQTT `PUBLISH <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718037>`_ message and wait for acknowledgment depending on Quality Of Service

            This method is a *coroutine*.

            :param topic: topic name to which message data is published
            :param message: payload message (as bytes) to send.
            :param qos: requested publish quality of service : QOS_0, QOS_1 or QOS_2. Defaults to ``default_qos`` config parameter or QOS_0.
            :param retain: retain flag. Defaults to ``default_retain`` config parameter or False.
            :param codec: Codec to encode the message with. Defaults to the connection's.
        """

        if codec is None:
            codec = self.codec
        elif isinstance(code,str):
            codec = _codecs[codec]()
        message = codec.encode(message)

        if qos is not None:
            assert qos in (QOS_0, QOS_1, QOS_2)
        else:
            try:
                qos = self.config['topics'][topic]['qos']
            except KeyError:
                qos = self.config['default_qos']
        if retain is None:
            try:
                retain = self.config['topics'][topic]['retain']
            except KeyError:
                retain = self.config['default_retain']
        return await self._handler.mqtt_publish(topic, message, qos, retain)

    @mqtt_connected
    async def subscribe(self, topics):
        """
            Subscribe to some topics.

            Send a MQTT `SUBSCRIBE <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718063>`_ message and wait for broker acknowledgment.

            This method is a *coroutine*.

            :param topics: array of topics pattern to subscribe with associated QoS.
            :return: `SUBACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718068>`_ message return code.

            Example of ``topics`` argument expected structure:
            ::

                [
                    ('$SYS/broker/uptime', QOS_1),
                    ('$SYS/broker/load/#', QOS_2),
                ]
        """
        return await self._handler.mqtt_subscribe(topics, self.session.next_packet_id)

    @mqtt_connected
    async def unsubscribe(self, topics):
        """
            Unsubscribe from some topics.

            Send a MQTT `UNSUBSCRIBE <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718072>`_ message and wait for broker `UNSUBACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718077>`_ message.

            This method is a *coroutine*.

            :param topics: array of topics to unsubscribe from.

            Example of ``topics`` argument expected structure:
            ::

                ['$SYS/broker/uptime', '$SYS/broker/load/#']
        """
        await self._handler.mqtt_unsubscribe(topics, self.session.next_packet_id)

    def subscription(self, topic, qos=QOS_0, codec=None):
        """
        Iterate over a message.

            :param topic: topic (including wildcards) to receive from.
            :param qos: minimum QOS to receive with.
            :param codec: Codec to decode incoming mssages with.

            Usage::

                async with client.subscription("/foo/bar") as sub:
                    async for msg in sub:
                        await process(msg)

        You can use either multiple calls to `subscription`, or manage
        message dispatching your self with `subscribe` and `deliver_message`.
        Using both methods in parallel are not supported.

        """
        class _Subscription:
            def __init__(self,client,topic,qos, codec=None):
                self.client = client
                self.topic = topic
                self.qos = qos
                if codec is None:
                    codec = client.codec
                elif isinstance(code,str):
                    codec = _codecs[codec]()
                self.codec = codec
                self._q = None

                global _handler_id
                _handler_id += 1
                self._id = _handler_id

            def __hash__(self):
                return self._id

            def __eq__(self, other):
                return self._id == getattr(other,'_id',other)

            async def __aenter__(self):
                self._q = anyio.create_queue(100)
                await self.client._subscribe(self)
                return self

            async def __aexit__(self, *tb):
                self._q = None
                async with anyio.open_cancel_scope(shield=True):
                    await self.client._unsubscribe(self)

            def __aiter__(self):
                return self

            async def __anext__(self):
                if self._q is None:
                    raise RuntimeError("Overflow. Please resubscribe.")
                message = await self._q.get()
                message.data = self.codec.decode(message.publish_packet.data)
                return message
            
            async def publish(self, topic, message, *a, **kw):
                """
                    Publish a message.

                    :param topic: topic name to which message data is published
                    :param message: payload message (as bytes) to send.
                    :param qos: requested publish quality of service : QOS_0, QOS_1 or QOS_2. Defaults to ``default_qos`` config parameter or QOS_0.
                    :param retain: retain flag. Defaults to ``default_retain`` config parameter or False.
                    :param codec: Codec to encode the message with. Defaults to the subscription's.
                """
                if topic is None:
                    topic = self.topic
                if 'codec' not in kw:
                    kw['codec'] = self.codec
                await self.client.publish(topic, message, *a, **kw)

        return _Subscription(self,topic,qos)

    async def _subscribe(self,handler):

        topic = handler.topic
        if self._subscriptions is None:
            self._subscriptions = dict()
            await self._tg.spawn(self._deliver_loop)
        clients = self._subscriptions.get(topic,None)
        if clients is None:
            self._subscriptions[topic] = clients = _set()
            clients.__topic = tuple(topic.split("/"))
            clients.__qos = handler.qos
            await self.subscribe([(topic,handler.qos)])
        elif clients.__qos < handler.qos:
            clients.__qos = handler.qos
            await self.subscribe([(topic,handler.qos)])

        clients.add(handler)

    async def _unsubscribe(self,handler):
        topic = handler.topic
        clients = self._subscriptions[topic]
        clients.remove(handler)
        if not clients:
            del self._subscriptions[topic]
            await self.unsubscribe([topic])

    async def _dispatch(self, msg):
        t = msg.topic.split('/')
        for topic,clients in self._subscriptions.items():
            if not match_topic(t, clients.__topic):
                continue
            for c in clients:
                if c._q is None:
                    continue
                try:
                    async with anyio.fail_after(0.01):
                        await c._q.put(msg)
                except TimeoutError:
                    c._q = None

    async def _deliver_loop(self):
        """
            Dispatch incoming messages to subscriptions.
        """
        async with anyio.open_cancel_scope() as scope:
            if self.client_task is not None:
                raise RuntimeError("You can't listen in more than one task")
            self.client_task = scope
            try:
                while True:
                    if self.session is None:
                        return
                    msg = await self.session.get_next_message()
                    if msg is None:
                        return
                    await self._dispatch(msg)

            finally:
                self.client_task = None


    async def deliver_message(self, codec=None):
        """
            Deliver next received message.

            Deliver next message received from the broker. If no message is available, this methods waits until next message arrives or ``timeout`` occurs.

            This method is a *coroutine*.

            :return: instance of :class:`distmqtt.session.ApplicationMessage` containing received message information flow.
            :raises: :class:`TimeoutError` if timeout occurs before a message is delivered

            :param codec: Codec to decode the message with.
            
            This method returns ``None`` if it is cancelled by closing the
            connection.
        """
        if codec is None:
            codec = self.codec
        elif isinstance(code,str):
            codec = _codecs[codec]()

        async with anyio.open_cancel_scope() as scope:
            if self.session is None:
                return None
            if self.client_task is not None:
                raise RuntimeError("You can't listen in more than one task")
            self.client_task = scope
            try:
                msg = await self.session.get_next_message()
                msg.data = codec.decode(msg.publish_packet.data)
                return msg

            finally:
                self.client_task = None

    async def _connect_coro(self):
        kwargs = dict()

        # Decode URI attributes
        uri_attributes = urlparse(self.session.broker_uri)
        scheme = uri_attributes.scheme
        secure = True if scheme in ('mqtts', 'wss') else False
        self.session.username = self.session.username if self.session.username else uri_attributes.username
        self.session.password = self.session.password if self.session.password else uri_attributes.password
        self.session.remote_address = uri_attributes.hostname
        self.session.remote_port = uri_attributes.port
        if scheme in ('mqtt', 'mqtts') and not self.session.remote_port:
            self.session.remote_port = 8883 if scheme == 'mqtts' else 1883
        if scheme in ('ws', 'wss') and not self.session.remote_port:
            self.session.remote_port = 443 if scheme == 'wss' else 80
        if scheme in ('ws', 'wss'):
            # Rewrite URI to conform to https://tools.ietf.org/html/rfc6455#section-3
            uri = (scheme, self.session.remote_address + ":" + str(self.session.remote_port), uri_attributes[2],
                   uri_attributes[3], uri_attributes[4], uri_attributes[5])
            self.session.broker_uri = urlunparse(uri)
        # Init protocol handler
        #if not self._handler:
        self._handler = ClientProtocolHandler(self.plugins_manager)

        if secure:
            sc = ssl.create_default_context(
                ssl.Purpose.SERVER_AUTH,
                cafile=self.session.cafile,
                capath=self.session.capath,
                cadata=self.session.cadata)
            if 'certfile' in self.config and 'keyfile' in self.config:
                sc.load_cert_chain(self.config['certfile'], self.config['keyfile'])
            if 'check_hostname' in self.config and isinstance(self.config['check_hostname'], bool):
                sc.check_hostname = self.config['check_hostname']
            kwargs['ssl_context'] = sc
            kwargs['autostart_tls'] = True

        try:
            adapter = None
            self._connected_state.clear()
            # Open connection
            if scheme in ('mqtt', 'mqtts'):
                conn = \
                    await anyio.connect_tcp(
                        self.session.remote_address,
                        self.session.remote_port, **kwargs)
                if secure:
                    await conn.start_tls()
                adapter = StreamAdapter(conn)
            elif scheme in ('ws', 'wss'):
                if kwargs.pop('autostart_tls', False):
                    kwargs['ssl'] = kwargs.pop('ssl_context')
                websocket = await create_websocket(self.session.broker_uri, 
                    subprotocols=['mqtt'],
                    headers=self.extra_headers,
                    **kwargs)
                adapter = WebSocketsAdapter(websocket)
            # Start MQTT protocol
            await self._handler.attach(self.session, adapter)
            try:
                return_code = await self._handler.mqtt_connect()
            except NoDataException:
                self.logger.warning("Connection broken by broker")
                exc = ConnectException("Connection broken by broker")
                raise exc
            if return_code is not CONNECTION_ACCEPTED:
                self.session.transitions.disconnect()
                self.logger.warning("Connection rejected with code '%s'", return_code)
                exc = ConnectException("Connection rejected by broker")
                exc.return_code = return_code
                raise exc
            else:
                # Handle MQTT protocol
                await self._handler.start()
                self.session.transitions.connect()
                await self._connected_state.set()
                self.logger.debug("connected to %s:%s", self.session.remote_address, self.session.remote_port)
            return return_code
        except ProtocolError as exc:
            self.logger.warning("connection failed: invalid websocket handshake")
            self.session.transitions.disconnect()
            raise ConnectException("connection failed: invalid websocket handshake") from exc
        except (ProtocolHandlerException, ConnectionError, OSError) as exc:
            self.logger.warning("MQTT connection failed")
            self.session.transitions.disconnect()
            raise ConnectException from exc

    async def handle_connection_close(self, evt):

        async def cancel_tasks():
            await self._no_more_connections.set()
            if self.client_task:
                task,self.client_task = self.client_task,None
                await task.cancel()

        async with anyio.open_cancel_scope() as scope:
            self._disconnect_task = scope
            await evt.set()
            self.logger.debug("Wait for broker disconnection")
            # Wait for disconnection from broker (like connection lost)
            await self._handler.wait_disconnect()
        self.logger.warning("Disconnected from broker")

        # Block client API
        self._connected_state.clear()

        # stop and clean handler
        await self._handler.stop()
        await self._handler.detach()
        self.session.transitions.disconnect()

        if self.config.get('auto_reconnect', False):
            # Try reconnection
            self.logger.debug("Auto-reconnecting")
            try:
                await self.reconnect()
            except ConnectException:
                # Cancel client pending tasks
                await cancel_tasks()
        else:
            # Cancel client pending tasks
            await cancel_tasks()

    def _initsession(
            self,
            uri=None,
            cleansession=None,
            cafile=None,
            capath=None,
            cadata=None) -> Session:
        # Load config
        broker_conf = self.config.get('broker', dict()).copy()
        if uri:
            broker_conf['uri'] = uri
        if cafile:
            broker_conf['cafile'] = cafile
        elif 'cafile' not in broker_conf:
            broker_conf['cafile'] = None
        if capath:
            broker_conf['capath'] = capath
        elif 'capath' not in broker_conf:
            broker_conf['capath'] = None
        if cadata:
            broker_conf['cadata'] = cadata
        elif 'cadata' not in broker_conf:
            broker_conf['cadata'] = None

        if cleansession is not None:
            broker_conf['cleansession'] = cleansession

        for key in ['uri']:
            if not_in_dict_or_none(broker_conf, key):
                raise ClientException("Missing connection parameter '%s'" % key)

        s = Session(self.plugins_manager)
        s.broker_uri = broker_conf['uri']
        s.client_id = self.client_id
        s.cafile = broker_conf['cafile']
        s.capath = broker_conf['capath']
        s.cadata = broker_conf['cadata']
        if cleansession is not None:
            s.clean_session = cleansession
        else:
            s.clean_session = self.config.get('cleansession', True)
        s.keep_alive = self.config['keep_alive'] - self.config['ping_delay']
        if 'will' in self.config:
            s.will_flag = True
            s.will_retain = self.config['will']['retain']
            s.will_topic = self.config['will']['topic']
            s.will_message = self.config['will']['message']
            s.will_qos = self.config['will']['qos']
        else:
            s.will_flag = False
            s.will_retain = False
            s.will_topic = None
            s.will_message = None
        return s



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest
import anyio
import trio
import os
import logging
import random
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

from distmqtt.client import open_mqttclient, ConnectException
from distmqtt.broker import create_broker
from distmqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from distkv.server import Server
from distkv.client import open_client
from functools import partial

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.INFO, format=formatter)
log = logging.getLogger(__name__)

broker_config = {
    'distkv': {
        'server': { "host": '127.0.0.1', "port": None },
        'serf': {'host':"localhost", 'port':7373},
        'topic': 'test_'+''.join(random.choices("abcdefghijklmnopqrstuvwxyz",k=7)),
        'retain': ('test','retain'),
    },
    'listeners': {
        'default': {
            'type': 'tcp',
            'bind': '127.0.0.1:1883',
            'max_connections': 10
        },
        'ws': {
            'type': 'ws',
            'bind': '127.0.0.1:8080',
            'max_connections': 10
        },
        'wss': {
            'type': 'ws',
            'bind': '127.0.0.1:8081',
            'max_connections': 10
        },
    },
    'sys_interval': 0,
    'auth': {
        'allow-anonymous': True,
    }
}


@asynccontextmanager
async def distkv_server(n):
    msgs = []
    async with anyio.create_task_group() as tg:
        s = Server("test", cfg={"serf":broker_config['distkv']['serf'], "server":{"host":"127.0.0.1", "port":None}}, init="test")
        evt = anyio.create_event()
        await tg.spawn(partial(s.serve, ready_evt=evt))
        await evt.wait()

        for h, p, *_ in s.ports:
            if h[0] != ":":
                break
        broker_config['distkv']['server']['host'] = h
        broker_config['distkv']['server']['port'] = p
        async with open_client(host=h,port=p) as cl:
            async def serflog(task_status=trio.TASK_STATUS_IGNORED):
                async with cl._stream('serfmon', type=broker_config['distkv']['topic']) as mon:
                    log.info("Serf Start")
                    task_status.started()
                    async for m in mon:
                        log.info("Serf Msg %r",m.data)
                        msgs.append(m.data)
            await cl.tg.spawn(serflog)
            yield s
            await cl.tg.cancel_scope.cancel()
        await tg.cancel_scope.cancel()
    assert len(msgs) == n, msgs


class MQTTClientTest(unittest.TestCase):
    def test_deliver(self):
        data = b'data 123'

        async def test_coro():
            async with distkv_server(1):
                async with create_broker(broker_config, plugin_namespace="distmqtt.test.plugins") as broker:
                    async with open_mqttclient() as client:
                        await client.connect('mqtt://127.0.0.1/')
                        self.assertIsNotNone(client.session)
                        ret = await client.subscribe([
                            ('test_topic', QOS_0),
                        ])
                        self.assertEqual(ret[0], QOS_0)
                        async with open_mqttclient() as client_pub:
                            await client_pub.connect('mqtt://127.0.0.1/')
                            await client_pub.publish('test_topic', data, QOS_0, retain=False)
                        #async with anyio.fail_after(0.5):
                        message = await client.deliver_message()
                        self.assertIsNotNone(message)
                        self.assertIsNotNone(message.publish_packet)
                        self.assertEqual(message.data, data)

        anyio.run(test_coro, backend='trio')

    def test_deliver_retain_direct(self):
        data = b'data 123'

        async def test_coro():
            async with distkv_server(0):
                async with create_broker(broker_config, plugin_namespace="distmqtt.test.plugins") as broker:
                    async with open_mqttclient() as client:
                        await client.connect('mqtt://127.0.0.1/')
                        self.assertIsNotNone(client.session)
                        ret = await client.subscribe([
                            ('test_topic', QOS_0),
                        ])
                        self.assertEqual(ret[0], QOS_0)
                        async with open_mqttclient() as client_pub:
                            await client_pub.connect('mqtt://127.0.0.1/')
                            await client_pub.publish('test_topic', data, QOS_0, retain=True)
                        async with anyio.fail_after(0.5):
                            message = await client.deliver_message()
                        self.assertIsNotNone(message)
                        self.assertIsNotNone(message.publish_packet)
                        self.assertEqual(message.data, data)

        anyio.run(test_coro, backend='trio')

    def test_deliver_retain_later(self):
        data = b'data 1234'

        async def test_coro():
            async with distkv_server(0) as s:
                async with create_broker(broker_config, plugin_namespace="distmqtt.test.plugins") as broker:
                    async with open_mqttclient() as client:
                        await client.connect('mqtt://127.0.0.1/')
                        self.assertIsNotNone(client.session)
                        async with open_mqttclient() as client_pub:
                            await client_pub.connect('mqtt://127.0.0.1/')
                            await client_pub.publish('test_topic', data, QOS_0, retain=True)
                    await anyio.sleep(1)

                    async with open_mqttclient() as client:
                        await client.connect('mqtt://127.0.0.1/')
                        self.assertIsNotNone(client.session)
                        ret = await client.subscribe([
                            ('test_topic', QOS_0),
                        ])
                        self.assertEqual(ret[0], QOS_0)
                        async with anyio.fail_after(0.5):
                            message = await client.deliver_message()
                        self.assertIsNotNone(message)
                        self.assertIsNotNone(message.publish_packet)
                        self.assertEqual(message.data, data)

                async with create_broker(broker_config, plugin_namespace="distmqtt.test.plugins") as broker:
                    async with open_mqttclient() as client:
                        await client.connect('mqtt://127.0.0.1/')
                        self.assertIsNotNone(client.session)
                        ret = await client.subscribe([
                            ('test_topic', QOS_0),
                        ])
                        self.assertEqual(ret[0], QOS_0)
                        async with anyio.fail_after(0.5):
                            message = await client.deliver_message()
                        self.assertIsNotNone(message)
                        self.assertIsNotNone(message.publish_packet)
                        self.assertEqual(message.data, data)

                seen = 0
                for h, p, *_ in s.ports:
                    if h[0] != ":":
                        break
                async with open_client(host=h,port=p) as cl:
                    async for m in cl.get_tree(min_depth=1):
                        del m['tock']
                        del m['seq']
                        assert m == {
                            'path': ('test', 'retain', 'test_topic'),
                            'value': b'data 1234'}
                        seen += 1
                assert seen == 1

        anyio.run(test_coro, backend='trio')

    def test_deliver_timeout(self):
        async def test_coro():
            async with distkv_server(0):
                async with create_broker(broker_config, plugin_namespace="distmqtt.test.plugins") as broker:
                    async with open_mqttclient() as client:
                        await client.connect('mqtt://127.0.0.1/')
                        self.assertIsNotNone(client.session)
                        ret = await client.subscribe([
                            ('test_topic', QOS_0),
                        ])
                        self.assertEqual(ret[0], QOS_0)
                        with self.assertRaises(TimeoutError):
                            async with anyio.fail_after(2):
                                await client.deliver_message()

        anyio.run(test_coro, backend='trio')



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import logging
import ssl
import asyncwebsockets
import anyio
import sys
import re
from collections import deque
try:
    from contextlib import asynccontextmanager
except ImportError:
    from async_generator import asynccontextmanager

from functools import partial
from transitions import Machine, MachineError
from distmqtt.session import Session, EVENT_BROKER_MESSAGE_RECEIVED
from distmqtt.mqtt.protocol.broker_handler import BrokerProtocolHandler
from distmqtt.errors import DistMQTTException, MQTTException
from distmqtt.utils import format_client_message, gen_client_id, Future, match_topic
from distmqtt.mqtt.constants import QOS_0
from distmqtt.adapters import (
    StreamAdapter,
    BaseAdapter,
    WebSocketsAdapter,
    )
from .plugins.manager import PluginManager, BaseContext
from asyncwebsockets import create_websocket_server


_defaults = {
    'listeners': {
        'default': {
            'type': 'tcp',
            'bind': '0.0.0.0:1883',
        },
    },
    'timeout-disconnect-delay': 2,
    'auth': {
        'allow-anonymous': True,
        'password-file': None
    },
}

EVENT_BROKER_PRE_START = 'broker_pre_start'
EVENT_BROKER_POST_START = 'broker_post_start'
EVENT_BROKER_PRE_SHUTDOWN = 'broker_pre_shutdown'
EVENT_BROKER_POST_SHUTDOWN = 'broker_post_shutdown'
EVENT_BROKER_CLIENT_CONNECTED = 'broker_client_connected'
EVENT_BROKER_CLIENT_DISCONNECTED = 'broker_client_disconnected'
EVENT_BROKER_CLIENT_SUBSCRIBED = 'broker_client_subscribed'
EVENT_BROKER_CLIENT_UNSUBSCRIBED = 'broker_client_unsubscribed'


class BrokerException(Exception):
    pass


class RetainedApplicationMessage:

    __slots__ = ('source_session', 'topic', 'data', 'qos')

    def __init__(self, source_session, topic, data, qos=None):
        self.source_session = source_session
        self.topic = topic
        self.data = data
        self.qos = qos


class Server:
    def __init__(self, listener_name, server_instance, max_connections=-1):
        self.logger = logging.getLogger(__name__)
        self.instance = server_instance
        self.conn_count = 0
        self.listener_name = listener_name

        self.max_connections = max_connections
        if self.max_connections > 0:
            self.semaphore = anyio.create_semaphore(self.max_connections)
        else:
            self.semaphore = None

    @asynccontextmanager
    async def _client_limit_(self):
        if self.semaphore:
            async with self.semaphore:
                yield self
        else:
            yield self

    @asynccontextmanager
    async def _client_limit(self):
        async with self._client_limit_():

            self.conn_count += 1
            if self.max_connections > 0:
                self.logger.info("Listener '%s': %d/%d connections acquired",
                                 self.listener_name, self.conn_count, self.max_connections)
            else:
                self.logger.info("Listener '%s': %d connections acquired",
                                 self.listener_name, self.conn_count)
            yield self

            self.conn_count -= 1
            if self.max_connections > 0:
                self.logger.info("Listener '%s': %d/%d connections acquired",
                                 self.listener_name, self.conn_count, self.max_connections)
            else:
                self.logger.info("Listener '%s': %d connections acquired",
                                 self.listener_name, self.conn_count)

    async def close_instance(self):
        await self.instance.cancel()


class BrokerContext(BaseContext):
    """
    BrokerContext is used as the context passed to plugins interacting with the broker.
    It act as an adapter to broker services from plugins developed for DistMQTT broker
    """
    def __init__(self, broker):
        super().__init__()
        self.config = None
        self._broker_instance = broker

    async def broadcast_message(self, topic, data, qos=None, retain=False):
        await self._broker_instance.internal_message_broadcast(topic, data, qos, retain=retain)

    @property
    def sessions(self):
        for k, session in self._broker_instance._sessions.items():
            yield session[0]

    @property
    def retained_messages(self):
        return self._broker_instance._retained_messages

    @property
    def subscriptions(self):
        return self._broker_instance._subscriptions


@asynccontextmanager
async def create_broker(config=None, plugin_namespace=None):
    """MQTT 3.1.1 compliant broker implementation 
    :param config: Example Yaml config
    :param plugin_namespace: Plugin namespace to use when loading plugin entry_points. Defaults to ``distmqtt.broker.plugins``

    This is an async context manager::

        async with create_broker() as broker:
            while True:
                anyio.sleep(99999)
    """

    async with anyio.create_task_group() as tg:
        if 'distkv' in (config or {}):
            from .distkv_broker import DistKVbroker
            B = DistKVbroker
        else:
            B = Broker
        b = B(tg, config, plugin_namespace)
        try:
            await b.start()
            yield b
        finally:
            await b.shutdown()
            await tg.cancel_scope.cancel()


class Broker:
    """
    MQTT 3.1.1 compliant broker implementation

    :param tg: The task group used to run the broker's tasks.
    :param config: Example Yaml config
    :param plugin_namespace: Plugin namespace to use when loading plugin entry_points. Defaults to ``distmqtt.broker.plugins``

    Usage::

        async with anyio.create_task_group() as tg: 
            b = Broker(tg, config, plugin_namespace)
            try:
                await b.start()
                pass ## do something with the broker
            finally:
                await b.shutdown()
                await tg.cancel_scope.cancel()

    Typically, though, you'll want to use :func:`create_broker`, which does
    this for you.
    """
    states = ['new', 'starting', 'started', 'not_started', 'stopping', 'stopped', 'not_stopped', 'stopped']

    def __init__(self, tg: anyio.abc.TaskGroup, config=None, plugin_namespace=None):
        self.logger = logging.getLogger(__name__)
        self.config = _defaults
        if config is not None:
            self.config.update(config)
        self._build_listeners_config(self.config)

        self._servers = dict()
        self._init_states()
        self._sessions = dict()
        self._subscriptions = dict()
        self._retained_messages = dict()
        self._broadcast_queue = anyio.create_queue(9999)
        self._tg = tg

        # Init plugins manager
        context = BrokerContext(self)
        context.config = self.config
        if plugin_namespace:
            namespace = plugin_namespace
        else:
            namespace = 'distmqtt.broker.plugins'
        self.plugins_manager = PluginManager(tg, namespace, context)

    def _build_listeners_config(self, broker_config):
        self.listeners_config = dict()
        try:
            listeners_config = broker_config['listeners']
            defaults = listeners_config.get('default', {})
            for listener in listeners_config:
                config = dict(defaults)
                config.update(listeners_config[listener])
                self.listeners_config[listener] = config
        except KeyError as ke:
            raise BrokerException("Listener config not found or invalid") from ke

    def _init_states(self):
        self.transitions = Machine(states=Broker.states, initial='new')
        self.transitions.add_transition(trigger='start', source='new', dest='starting')
        self.transitions.add_transition(trigger='starting_fail', source='starting', dest='not_started')
        self.transitions.add_transition(trigger='starting_success', source='starting', dest='started')
        self.transitions.add_transition(trigger='shutdown', source='starting', dest='not_started')
        self.transitions.add_transition(trigger='shutdown', source='started', dest='stopping')
        self.transitions.add_transition(trigger='shutdown', source='not_started', dest='stopping')
        self.transitions.add_transition(trigger='stopping_success', source='stopped', dest='stopped')
        self.transitions.add_transition(trigger='stopping_success', source='not_started', dest='not_started')
        self.transitions.add_transition(trigger='stopping_failure', source='stopping', dest='not_stopped')
        self.transitions.add_transition(trigger='start', source='stopped', dest='starting')
        self.transitions.add_transition(trigger='shutdown', source='new', dest='stopped')
        self.transitions.add_transition(trigger='stopping_success', source='stopping', dest='stopped')

    async def start(self):
        """
            Start the broker to serve with the given configuration

            Start method opens network sockets and will start listening for incoming connections.

            This method is a *coroutine*.
        """
        try:
            self._sessions = dict()
            self._subscriptions = dict()
            self._retained_messages = dict()
            self.transitions.start()
            self.logger.debug("Broker starting")
        except (MachineError, ValueError) as exc:
            # Backwards compat: MachineError is raised by transitions < 0.5.0.
            self.logger.warning("[WARN-0001] Invalid method call at this moment: %r", exc)
            raise BrokerException("Broker instance can't be started: %s" % exc)

        await self.plugins_manager.fire_event(EVENT_BROKER_PRE_START)
        try:
            # Start network listeners
            for listener_name in self.listeners_config:
                listener = self.listeners_config[listener_name]

                if 'bind' not in listener:
                    self.logger.debug("Listener configuration '%s' is not bound", listener_name)
                else:
                    # Max connections
                    try:
                        max_connections = listener['max_connections']
                    except KeyError:
                        max_connections = -1

                    # SSL Context
                    sc = None

                    # accept string "on" / "off" or boolean
                    ssl_active = listener.get('ssl', False)
                    if isinstance(ssl_active, str):
                        ssl_active = ssl_active.upper() == 'ON'

                    if ssl_active:
                        try:
                            sc = ssl.create_default_context(
                                ssl.Purpose.CLIENT_AUTH,
                                cafile=listener.get('cafile'),
                                capath=listener.get('capath'),
                                cadata=listener.get('cadata')
                            )
                            sc.load_cert_chain(listener['certfile'], listener['keyfile'])
                            sc.verify_mode = ssl.CERT_OPTIONAL
                        except KeyError as ke:
                            raise BrokerException("'certfile' or 'keyfile' configuration parameter missing: %s" % ke)
                        except FileNotFoundError as fnfe:
                            raise BrokerException("Can't read cert files '%s' or '%s' : %s" %
                                                  (listener['certfile'], listener['keyfile'], fnfe))

                    address, s_port = listener['bind'].split(':')
                    port = 0
                    try:
                        port = int(s_port)
                    except ValueError as ve:
                        raise BrokerException("Invalid port value in bind value: %s" % listener['bind'])

                    async def server_task(evt, cb, address,port,ssl):
                        async with anyio.open_cancel_scope() as scope:
                            await evt.set(scope)
                            async with await anyio.create_tcp_server(port, interface=address, ssl_context=ssl) as server:
                                async for conn in server.accept_connections():
                                    await self._tg.spawn(cb, conn)

                    if listener['type'] == 'tcp':
                        cb_partial = partial(self.stream_connected, listener_name=listener_name)
                    elif listener['type'] == 'ws':
                        cb_partial = partial(self.ws_connected, listener_name=listener_name)
                    else:
                        self.logger.error("Listener '%s': unknown type '%s'", listener_name, listener['type'])
                        continue
                    fut = Future()
                    await self._tg.spawn(server_task, fut, cb_partial, address,port,sc,name=listener_name)
                    instance = await fut.get()
                    self._servers[listener_name] = Server(listener_name, instance, max_connections)

                    self.logger.info("Listener '%s' bind to %s (max_connections=%d)",
                                     listener_name, listener['bind'], max_connections)

            self.transitions.starting_success()
            await self.plugins_manager.fire_event(EVENT_BROKER_POST_START)

            #Start broadcast loop
            await self._tg.spawn(self._broadcast_loop)

            self.logger.debug("Broker started")
        except Exception as e:
            if "Cancel" in repr(e):
                raise  # bah
            self.logger.error("Broker startup failed: %r", e)
            self.transitions.starting_fail()
            raise BrokerException("Broker instance can't be started") from e

    async def shutdown(self):
        """
            Stop broker instance.

            Closes all connected session, stop listening on network socket and free resources.
        """
        for s in self._sessions.values():
            await s[0].stop()

        self._sessions = dict()
        self._subscriptions = dict()
        self._retained_messages = dict()
        try:
            self.transitions.shutdown()
        except MachineError as exc:
            # Backwards compat: MachineError is raised by transitions < 0.5.0.
            raise BrokerException("Broker instance can't be stopped") from exc

        # Fire broker_shutdown event to plugins
        await self.plugins_manager.fire_event(EVENT_BROKER_PRE_SHUTDOWN)

        # Stop broadcast loop
        if self._broadcast_queue.qsize() > 0:
            self.logger.warning("%d messages not broadcasted", self._broadcast_queue.qsize())

        for listener_name in self._servers:
            server = self._servers[listener_name]
            await server.close_instance()
        self.logger.debug("Broker closing")
        self.logger.info("Broker closed")
        await self.plugins_manager.fire_event(EVENT_BROKER_POST_SHUTDOWN)
        self.transitions.stopping_success()

    async def internal_message_broadcast(self, topic, data, qos=None, retain=None):
        return await self.broadcast_message(None, topic, data, qos=qos, retain=retain)

    async def ws_connected(self, conn, listener_name):
        async def subpro(req):
            if "mqtt" not in req.subprotocols:
                return False
            return "mqtt"
        websocket = await create_websocket_server(conn, filter=subpro)
        await self.client_connected(listener_name, WebSocketsAdapter(websocket))

    async def stream_connected(self, conn, listener_name):
        await self.client_connected(listener_name, StreamAdapter(conn))

    async def client_connected(self, listener_name, adapter: BaseAdapter):
        server = self._servers.get(listener_name, None)
        if not server:
            raise BrokerException("Invalid listener name '%s'" % listener_name)

        async with server._client_limit():
            return await self.client_connected_(server, listener_name, adapter)

    async def client_connected_(self, server, listener_name, adapter: BaseAdapter):
        # Wait for connection available on listener

        remote_address, remote_port = adapter.get_peer_info()
        self.logger.info("Connection from %s:%d on listener '%s'", remote_address, remote_port, listener_name)

        # Wait for first packet and expect a CONNECT
        try:
            handler, client_session = await BrokerProtocolHandler.init_from_connect(adapter, self.plugins_manager)
        except DistMQTTException as exc:
            self.logger.warning("[MQTT-3.1.0-1] %s: Can't read first packet an CONNECT: %s",
                                format_client_message(address=remote_address, port=remote_port), exc)
            #await writer.close()
            self.logger.debug("Connection closed")
            return
        except MQTTException as me:
            self.logger.error('Invalid connection from %s : %s',
                              format_client_message(address=remote_address, port=remote_port), me)
            await adapter.close()
            self.logger.debug("Connection closed")
            return

        if client_session.clean_session:
            # Delete existing session and create a new one
            if client_session.client_id is not None and client_session.client_id != "":
                await self.delete_session(client_session.client_id)
            else:
                client_session.client_id = gen_client_id()
            client_session.parent = 0
        else:
            # Get session from cache
            if client_session.client_id in self._sessions:
                self.logger.debug("Found old session %r", self._sessions[client_session.client_id])
                (client_session, h) = self._sessions[client_session.client_id]
                client_session.parent = 1
            else:
                client_session.parent = 0
        if not client_session.parent:
            await client_session.start(self)
        if client_session.keep_alive > 0 and not client_session.parent:
            # MQTT 3.1.2.10: one and a half keepalive times, plus configurable grace
            client_session.keep_alive += client_session.keep_alive/2 + self.config['timeout-disconnect-delay']
        self.logger.debug("Keep-alive timeout=%d", client_session.keep_alive)

        await handler.attach(client_session, adapter)
        self._sessions[client_session.client_id] = (client_session, handler)

        authenticated = await self.authenticate(client_session, self.listeners_config[listener_name])
        if not authenticated:
            await adapter.close()
            return

        while True:
            try:
                client_session.transitions.connect()
                break
            except (MachineError, ValueError) as exc:
                # Backwards compat: MachineError is raised by transitions < 0.5.0.
                self.logger.warning("Client %s is reconnecting too quickly, make it wait", client_session.client_id, exc_info=exc)
                # Wait a bit may be client is reconnecting too fast
                await anyio.sleep(1)
        await handler.mqtt_connack_authorize(authenticated)

        await self.plugins_manager.fire_event(EVENT_BROKER_CLIENT_CONNECTED, client_id=client_session.client_id)

        self.logger.debug("%s Start messages handling", client_session.client_id)
        await handler.start()
        self.logger.debug("Retained messages queue size: %d", client_session.retained_messages.qsize())
        await self.publish_session_retained_messages(client_session)

        # Init and start loop for handling client messages (publish, subscribe/unsubscribe, disconnect)
        async with anyio.create_task_group() as tg:

            async def handle_unsubscribe():
                while True:
                    unsubscription = await handler.get_next_pending_unsubscription()
                    self.logger.debug("%s handling unsubscription", client_session.client_id)
                    for topic in unsubscription['topics']:
                        self._del_subscription(topic, client_session)
                        await self.plugins_manager.fire_event(
                            EVENT_BROKER_CLIENT_UNSUBSCRIBED,
                            client_id=client_session.client_id,
                            topic=topic)
                    await handler.mqtt_acknowledge_unsubscription(unsubscription['packet_id'])

            async def handle_subscribe():
                while True:
                    subscriptions = await handler.get_next_pending_subscription()
                    self.logger.debug("%s handling subscription", client_session.client_id)
                    return_codes = []
                    for subscription in subscriptions['topics']:
                        result = await self.add_subscription(subscription, client_session)
                        return_codes.append(result)
                    await handler.mqtt_acknowledge_subscription(subscriptions['packet_id'], return_codes)
                    for index, subscription in enumerate(subscriptions['topics']):
                        if return_codes[index] != 0x80:
                            await self.plugins_manager.fire_event(
                                EVENT_BROKER_CLIENT_SUBSCRIBED,
                                client_id=client_session.client_id,
                                topic=subscription[0],
                                qos=subscription[1])
                            await self.publish_retained_messages_for_subscription(subscription, client_session)
                    self.logger.debug(repr(self._subscriptions))

            await tg.spawn(handle_unsubscribe)
            await tg.spawn(handle_subscribe)

            try:
                await handler.wait_disconnect()
                self.logger.debug("%s wait_diconnect: %sclean", client_session.client_id, "" if handler.clean_disconnect else "un")

                if not handler.clean_disconnect:
                    # Connection closed anormally, send will message
                    self.logger.debug("Will flag: %s", client_session.will_flag)
                    if client_session.will_flag:
                        if self.logger.isEnabledFor(logging.DEBUG):
                            self.logger.debug("Client %s disconnected abnormally, sending will message",
                                              format_client_message(session=client_session))
                        await self.broadcast_message(
                            client_session,
                            client_session.will_topic,
                            client_session.will_message,
                            client_session.will_qos,
                            retain=client_session.will_retain)
                self.logger.debug("%s Disconnecting session", client_session.client_id)
                await self._stop_handler(handler)
                client_session.transitions.disconnect()
                await self.plugins_manager.fire_event(EVENT_BROKER_CLIENT_DISCONNECTED, client_id=client_session.client_id)
            finally:
                async with anyio.open_cancel_scope(shield=True):
                    await tg.cancel_scope.cancel()
            pass # end taskgroup

        self.logger.debug("%s Client disconnected", client_session.client_id)

    async def _stop_handler(self, handler):
        """
        Stop a running handler and detach if from the session
        :param handler:
        :return:
        """
        try:
            await handler.detach()
            await handler.stop()
        except Exception as e:
            #self.logger.exception("Stop handler")
            raise

    async def authenticate(self, session: Session, listener):
        """
        This method call the authenticate method on registered plugins to test user authentication.
        User is considered authenticated if all plugins called returns True.
        Plugins authenticate() method are supposed to return :
         - True if user is authentication succeed
         - False if user authentication fails
         - None if authentication can't be achieved (then plugin result is then ignored)
        :param session:
        :param listener:
        :return:
        """
        auth_plugins = None
        auth_config = self.config.get('auth', None)
        if auth_config:
            auth_plugins = auth_config.get('plugins', None)
        returns = await self.plugins_manager.map_plugin_coro(
            "authenticate",
            session=session,
            filter_plugins=auth_plugins)
        auth_result = True
        if returns:
            for plugin in returns:
                res = returns[plugin]
                if res is False:
                    auth_result = False
                    self.logger.debug("Authentication failed due to '%s' plugin result: %s", plugin.name, res)
                else:
                    self.logger.debug("'%s' plugin result: %s", plugin.name, res)
        # If all plugins returned True, authentication is success
        return auth_result

    async def topic_filtering(self, session: Session, topic):
        """
        This method call the topic_filtering method on registered plugins to check that the subscription is allowed.
        User is considered allowed if all plugins called return True.
        Plugins topic_filtering() method are supposed to return :
         - True if MQTT client can be subscribed to the topic
         - False if MQTT client is not allowed to subscribe to the topic
         - None if topic filtering can't be achieved (then plugin result is then ignored)
        :param session:
        :param listener:
        :param topic: Topic in which the client wants to subscribe
        :return:
        """
        topic_plugins = None
        topic_config = self.config.get('topic-check', None)
        if topic_config and topic_config.get('enabled', False):
            topic_plugins = topic_config.get('plugins', None)
        returns = await self.plugins_manager.map_plugin_coro(
            "topic_filtering",
            session=session,
            topic=topic,
            filter_plugins=topic_plugins)

        topic_result = True
        if returns:
            for plugin in returns:
                res = returns[plugin]
                if res is False:
                    topic_result = False
                    self.logger.debug("Topic filtering failed due to '%s' plugin result: %s", plugin.name, res)
                else:
                    self.logger.debug("'%s' plugin result: %s", plugin.name, res)
        # If all plugins returned True, authentication is success
        return topic_result

    def retain_message(self, source_session, topic_name, data, qos=None):
        if data is not None and data != b'':
            # If retained flag set, store the message for further subscriptions
            self.logger.debug("Retaining message on topic %s", topic_name)
            retained_message = RetainedApplicationMessage(source_session, topic_name, data, qos)
            self._retained_messages[topic_name] = retained_message
        else:
            # [MQTT-3.3.1-10]
            if topic_name in self._retained_messages:
                self.logger.debug("Clear retained messages for topic '%s'", topic_name)
                del self._retained_messages[topic_name]

    async def add_subscription(self, subscription, session):
        a_filter = subscription[0]
        if '#' in a_filter and not a_filter.endswith('#'):
            # [MQTT-4.7.1-2] Wildcard character '#' is only allowed as last character in filter
            return 0x80
        if a_filter != "+":
            if '+' in a_filter:
                if "/+" not in a_filter and "+/" not in a_filter:
                    # [MQTT-4.7.1-3] + wildcard character must occupy entire level
                    return 0x80

        # Check if the client is authorised to connect to the topic
        permitted = await self.topic_filtering(session, topic=a_filter)
        if not permitted:
            return 0x80

        a_filter = tuple(a_filter.split('/'))
        qos = subscription[1]
        if 'max-qos' in self.config and qos > self.config['max-qos']:
            qos = self.config['max-qos']
        if a_filter not in self._subscriptions:
            self._subscriptions[a_filter] = []
        already_subscribed = next(
            (s for (s, qos) in self._subscriptions[a_filter] if s.client_id == session.client_id), None)
        if not already_subscribed:
            self._subscriptions[a_filter].append((session, qos))
        else:
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.debug("Client %s has already subscribed to %s", format_client_message(session=session), a_filter)
        return qos

    def _del_subscription(self, a_filter, session):
        """
        Delete a session subscription on a given topic
        :param a_filter:
        :param session:
        :return:
        """
        deleted = 0
        if isinstance(a_filter,str):
            a_filter = tuple(a_filter.split('/'))
        try:
            subscriptions = self._subscriptions[a_filter]
            for index, (sub_session, qos) in enumerate(subscriptions):
                if sub_session.client_id == session.client_id:
                    if self.logger.isEnabledFor(logging.DEBUG):
                        self.logger.debug("Removing subscription on topic '%s' for client %s",
                                          a_filter, format_client_message(session=session))
                    subscriptions.pop(index)
                    deleted += 1
                    break
        except KeyError:
            # Unsubscribe topic not found in current subscribed topics
            pass
        finally:
            return deleted

    def _del_all_subscriptions(self, session):
        """
        Delete all topic subscriptions for a given session
        :param session:
        :return:
        """
        filter_queue = deque()
        for topic in self._subscriptions:
            if self._del_subscription(topic, session):
                filter_queue.append(topic)
        for topic in filter_queue:
            if not self._subscriptions[topic]:
                del self._subscriptions[topic]

    async def _broadcast_loop(self):
        async with anyio.create_task_group() as tg:
            while True:
                broadcast = await self._broadcast_queue.get()
                self.logger.debug("broadcasting %r", broadcast)
                topic = broadcast['topic'].split('/')

                targets = {}
                for k_filter,subscriptions in self._subscriptions.items():
                    if match_topic(topic, k_filter):
                        for (target_session, qos) in subscriptions:
                            qos = max(qos, broadcast.get('qos', QOS_0), targets.get(target_session, QOS_0))
                            targets[target_session] = qos

                for target_session, qos in targets.items():
                    if target_session.transitions.state == 'connected':
                        if self.logger.isEnabledFor(logging.DEBUG):
                            self.logger.debug("broadcasting application message from %s on topic '%s' to %s",
                                                format_client_message(session=broadcast['session']),
                                                broadcast['topic'], format_client_message(session=target_session))
                        handler = self._get_handler(target_session)
                        await tg.spawn(partial(handler.mqtt_publish,broadcast['topic'], broadcast['data'], qos, retain=False))
                    else:
                        if self.logger.isEnabledFor(logging.DEBUG):
                            self.logger.debug("retaining application message from %s on topic '%s' to client '%s'",
                                                format_client_message(session=broadcast['session']),
                                                broadcast['topic'], format_client_message(session=target_session))
                        retained_message = RetainedApplicationMessage(
                            broadcast['session'], broadcast['topic'], broadcast['data'], qos)
                        await target_session.retained_messages.put(retained_message)

    async def broadcast_message(self, session, topic, data, force_qos=None, qos=None, retain=False):
        broadcast = {
            'session': session,
            'topic': topic,
            'data': data,
        }
        if force_qos:
            broadcast['qos'] = force_qos
        await self._broadcast_queue.put(broadcast)
        if retain:
            self.retain_message(session, topic, data, force_qos or qos)

    async def publish_session_retained_messages(self, session):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug("Publishing %d messages retained for session %s",
                              session.retained_messages.qsize(), format_client_message(session=session))
        handler = self._get_handler(session)
        async with anyio.create_task_group() as tg:
            while not session.retained_messages.empty():
                retained = await session.retained_messages.get()
                await tg.spawn(handler.mqtt_publish,
                    retained.topic, retained.data, retained.qos, True)

    async def publish_retained_messages_for_subscription(self, subscription, session):
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug("Begin broadcasting messages retained due to subscription on '%s' from %s",
                              subscription[0], format_client_message(session=session))
        handler = self._get_handler(session)
        async with anyio.create_task_group() as tg:
            for d_topic in self._retained_messages:
                topic = d_topic.split('/')
                self.logger.debug("matching : %s %s", d_topic, subscription[0])
                if match_topic(topic, subscription[0]):
                    self.logger.debug("%s and %s match", d_topic, subscription[0])
                    retained = self._retained_messages[d_topic]
                    await tg.spawn(handler.mqtt_publish,
                            retained.topic, retained.data, subscription[1], True)
        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug("End broadcasting messages retained due to subscription on '%s' from %s",
                              subscription[0], format_client_message(session=session))

    async def delete_session(self, client_id):
        """
        Delete an existing session data, for example due to clean session set in CONNECT
        :param client_id:
        :return:
        """
        try:
            session = self._sessions[client_id][0]
        except KeyError:
            session = None
        if session is None:
            self.logger.debug("Delete session : session %s doesn't exist", client_id)
            return

        # Delete subscriptions
        self.logger.debug("deleting session %r subscriptions", session)
        self._del_all_subscriptions(session)

        self.logger.debug("deleting existing session %r", self._sessions[client_id])
        await session.stop()
        del self._sessions[client_id]

    def _get_handler(self, session):
        client_id = session.client_id
        if client_id:
            try:
                return self._sessions[client_id][1]
            except KeyError:
                pass
        return None



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import sqlite3
import pickle


class SQLitePlugin:
    def __init__(self, context):
        self.context = context
        self.conn = None
        self.cursor = None
        self.db_file = None
        try:
            self.persistence_config = self.context.config['persistence']
            self.init_db()
        except KeyError:
            self.context.logger.warning("'persistence' section not found in context configuration")

    def init_db(self):
        self.db_file = self.persistence_config.get('file', None)
        if not self.db_file:
            self.context.logger.warning("'file' persistence parameter not found")
        else:
            try:
                self.conn = sqlite3.connect(self.db_file)
                self.cursor = self.conn.cursor()
                self.context.logger.info("Database file '%s' opened", self.db_file)
            except Exception as e:
                self.context.logger.error("Error while initializing database '%s' : %s", self.db_file, e)
                raise
        if self.cursor:
            self.cursor.execute("CREATE TABLE IF NOT EXISTS session(client_id TEXT PRIMARY KEY, data BLOB)")

    async def save_session(self, session):
        if self.cursor:
            dump = pickle.dumps(session)
            try:
                self.cursor.execute(
                    "INSERT OR REPLACE INTO session (client_id, data) VALUES (?,?)", (session.client_id, dump))
                self.conn.commit()
            except Exception as e:
                self.context.logger.error("Failed saving session '%s': %s", session, e)
                raise

    async def find_session(self, client_id):
        if self.cursor:
            row = self.cursor.execute("SELECT data FROM session where client_id=?", (client_id,)).fetchone()
            if row:
                return pickle.loads(row[0])
            else:
                return None

    async def del_session(self, client_id):
        if self.cursor:
            self.cursor.execute("DELETE FROM session where client_id=?", (client_id,))
            self.conn.commit()

    async def on_broker_post_shutdown(self):
        if self.conn:
            self.conn.close()
            self.context.logger.info("Database file '%s' closed", self.db_file)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import anyio
from struct import pack, unpack
from distmqtt.errors import NoDataException

# Collection of basic codecs for the messages' payload

import msgpack

def bytes_to_hex_str(data):
    """
    converts a sequence of bytes into its displayable hex representation, ie: 0x??????
    :param data: byte sequence
    :return: Hexadecimal displayable representation
    """
    return '0x' + ''.join(format(b, '02x') for b in data)


def bytes_to_int(data):
    """
    convert a sequence of bytes to an integer using big endian byte ordering
    :param data: byte sequence
    :return: integer value
    """
    try:
        return int.from_bytes(data, byteorder='big')
    except:
        return data


def int_to_bytes(int_value: int, length: int) -> bytes:
    """
    convert an integer to a sequence of bytes using big endian byte ordering
    :param int_value: integer value to convert
    :param length: (optional) byte length
    :return: byte sequence
    """
    if length == 1:
        fmt = "!B"
    elif length == 2:
        fmt = "!H"
    return pack(fmt, int_value)


async def read_or_raise(reader, n=-1):
    """
    Read a given byte number from Stream. NoDataException is raised if read gives no data
    :param reader: reader adapter
    :param n: number of bytes to read
    :return: bytes read
    """
    try:
        data = await reader.read(n)
    except (anyio.exceptions.IncompleteRead, ConnectionResetError):
        data = None
    if not data:
        raise NoDataException("No more data")
    return data


async def decode_string(reader) -> bytes:
    """
    Read a string from a reader and decode it according to MQTT string specification
    :param reader: Stream reader
    :return: UTF-8 string read from stream
    """
    length_bytes = await read_or_raise(reader, 2)
    str_length = unpack("!H", length_bytes)
    if str_length[0]:
        byte_str = await read_or_raise(reader, str_length[0])
        try:
            return byte_str.decode(encoding='utf-8')
        except:
            return str(byte_str)
    else:
        return ''


async def decode_data_with_length(reader) -> bytes:
    """
    Read data from a reader. Data is prefixed with 2 bytes length
    :param reader: Stream reader
    :return: bytes read from stream (without length)
    """
    length_bytes = await read_or_raise(reader, 2)
    bytes_length = unpack("!H", length_bytes)
    data = await read_or_raise(reader, bytes_length[0])
    return data


def encode_string(string: str) -> bytes:
    data = string.encode(encoding='utf-8')
    data_length = len(data)
    return int_to_bytes(data_length, 2) + data


def encode_data_with_length(data: bytes) -> bytes:
    data_length = len(data)
    return int_to_bytes(data_length, 2) + data


async def decode_packet_id(reader) -> int:
    """
    Read a packet ID as 2-bytes int from stream according to MQTT specification (2.3.1)
    :param reader: Stream reader
    :return: Packet ID
    """
    packet_id_bytes = await read_or_raise(reader, 2)
    packet_id = unpack("!H", packet_id_bytes)
    return packet_id[0]


def int_to_bytes_str(value: int) -> bytes:
    """
    Converts a int value to a bytes array containing the numeric character.
    Ex: 123 -> b'123'
    :param value: int value to convert
    :return: bytes array
    """
    return str(value).encode('utf-8')



class NoopCodec:
    """A codec that does nothing.

    Your payload needs to consist of bytes.
    """

    @staticmethod
    def encode(data):
        assert isinstance(data, bytes)
        return data

    @staticmethod
    def decode(data):
        return data


class UTF8Codec:
    """A codec that translates to UTF-8 strings.

    Your payload needs to be a single string.

    This codec will *not* stringify other data types for you.
    """

    @staticmethod
    def encode(data):
        return data.encode("utf-8")

    @staticmethod
    def decode(data):
        return data.decode("utf-8")


class MsgPackCodec:
    """A codec that encodes to "msgpack"-encoded bytestring.

    Your payload must consist of whatever "msgpack" accepts.

    Args:
      ``use_bin_type``: Bytestrings are encoded as byte arrays.
                        Defaults to ``True``. If ``False``,
                        bytes are transmitted as string types and input
                        strings are not UTF8-decoded. Use this if your
                        network contains clients which use
                        non-UTF8-encoded strings (this violates the
                        MsgPack specification).
      ``use_list``: if ``True``, lists and tuples are returned as lists
                    (i.e. modifyable). Defaults to ``False``, which uses
                    immutable tuples (this is faster).
    """

    use_bin_type = True
    use_list = False

    def __init__(self, use_bin_type=None, use_list=None):
        if use_bin_type is not None:
            self.use_bin_type = use_bin_type
        if use_list is not None:
            self.use_list = use_list

    def encode(self, data):
        return msgpack.packb(data, use_bin_type=self.use_bin_type)

    def decode(self, data):
        return msgpack.unpackb(data, raw=not self.use_bin_type, use_list=self.use_list)
        # raw=False would try to decode input bytes, which is not what you
        # want when the input is a bytestring. So we only use that if
        # bytestring input has been marked as such.




# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import anyio
import logging

from transitions import Machine
from collections import OrderedDict

from distmqtt.mqtt.publish import PublishPacket
from distmqtt.errors import DistMQTTException, MQTTException
from distmqtt.mqtt.constants import QOS_0

OUTGOING = 0
INCOMING = 1

EVENT_BROKER_MESSAGE_RECEIVED = 'broker_message_received'


class ApplicationMessage:

    """
        ApplicationMessage and subclasses are used to store published message information flow. These objects can contain different information depending on the way they were created (incoming or outgoing) and the quality of service used between peers.
    """

    __slots__ = (
        'packet_id', 'topic', 'qos', 'data', 'retain', 'publish_packet',
        'puback_packet', 'pubrec_packet', 'pubrel_packet', 'pubcomp_packet',
    )

    def __init__(self, packet_id, topic, qos, data, retain):
        self.packet_id = packet_id
        """ Publish message `packet identifier <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718025>`_"""

        self.topic = topic
        """ Publish message topic"""

        self.qos = qos
        """ Publish message Quality of Service"""

        self.data = data
        """ Publish message payload data"""

        self.retain = retain
        """ Publish message retain flag"""

        self.publish_packet = None
        """ :class:`distmqtt.mqtt.publish.PublishPacket` instance corresponding to the `PUBLISH <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718037>`_ packet in the messages flow. ``None`` if the PUBLISH packet has not already been received or sent."""

        self.puback_packet = None
        """ :class:`distmqtt.mqtt.puback.PubackPacket` instance corresponding to the `PUBACK <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718043>`_ packet in the messages flow. ``None`` if QoS != QOS_1 or if the PUBACK packet has not already been received or sent."""

        self.pubrec_packet = None
        """ :class:`distmqtt.mqtt.puback.PubrecPacket` instance corresponding to the `PUBREC <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718048>`_ packet in the messages flow. ``None`` if QoS != QOS_2 or if the PUBREC packet has not already been received or sent."""

        self.pubrel_packet = None
        """ :class:`distmqtt.mqtt.puback.PubrelPacket` instance corresponding to the `PUBREL <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718053>`_ packet in the messages flow. ``None`` if QoS != QOS_2 or if the PUBREL packet has not already been received or sent."""

        self.pubcomp_packet = None
        """ :class:`distmqtt.mqtt.puback.PubrelPacket` instance corresponding to the `PUBCOMP <http://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html#_Toc398718058>`_ packet in the messages flow. ``None`` if QoS != QOS_2 or if the PUBCOMP packet has not already been received or sent."""

    def build_publish_packet(self, dup=False):
        """
            Build :class:`distmqtt.mqtt.publish.PublishPacket` from attributes

        :param dup: force dup flag
        :return: :class:`distmqtt.mqtt.publish.PublishPacket` built from ApplicationMessage instance attributes
        """
        return PublishPacket.build(self.topic, self.data, self.packet_id, dup, self.qos, self.retain)

    def __eq__(self, other):
        return self.packet_id == other.packet_id

    def __getstate__(self):
        res = dict((k,getattr(self,k)) for k in ('topic','qos','data','retain'))
        res['id'] = self.packet_id
        return res

    def __setstate__(self, state):
        self.packet_id = res['id']
        for k in ('topic','qos','data','retain'):
            setattr(self, k, state[k])

    
class IncomingApplicationMessage(ApplicationMessage):

    """
        Incoming :class:`~distmqtt.session.ApplicationMessage`.
    """

    __slots__ = ('direction',)

    def __init__(self, packet_id, topic, qos, data, retain):
        super().__init__(packet_id, topic, qos, data, retain)
        self.direction = INCOMING


class OutgoingApplicationMessage(ApplicationMessage):

    """
        Outgoing :class:`~distmqtt.session.ApplicationMessage`.
    """

    __slots__ = ('direction',)

    def __init__(self, packet_id, topic, qos, data, retain):
        super().__init__(packet_id, topic, qos, data, retain)
        self.direction = OUTGOING


class Session:
    states = ['new', 'connected', 'disconnected']

    def __init__(self, plugins_manager):
        self._init_states()
        self._plugins_manager = plugins_manager
        self.remote_address = None
        self.remote_port = None
        self.client_id = None
        self.clean_session = None
        self.will_flag = False
        self.will_message = None
        self.will_qos = None
        self.will_retain = None
        self.will_topic = None
        self.keep_alive = 0
        self.publish_retry_delay = 0
        self.broker_uri = None
        self.username = None
        self.password = None
        self.cafile = None
        self.capath = None
        self.cadata = None
        self._packet_id = 0
        self.parent = 0

        self.logger = logging.getLogger(__name__)

        # Used to store outgoing ApplicationMessage while publish protocol flows
        self.inflight_out = OrderedDict()

        # Used to store incoming ApplicationMessage while publish protocol flows
        self.inflight_in = OrderedDict()

        # Stores messages retained for this session
        self.retained_messages = anyio.create_queue(9999)

        # Stores PUBLISH messages ID received in order and ready for application process
        self._delivered_message_queue = anyio.create_queue(9999)

        # The actual delivery process
        self._delivery_task = None
        self._delivery_stopped = anyio.create_event()

        # The broker we're attached to
        self._broker = None

    def _init_states(self):
        self.transitions = Machine(states=Session.states, initial='new')
        self.transitions.add_transition(trigger='connect', source='new', dest='connected')
        self.transitions.add_transition(trigger='connect', source='disconnected', dest='connected')
        self.transitions.add_transition(trigger='disconnect', source='connected', dest='disconnected')
        self.transitions.add_transition(trigger='disconnect', source='new', dest='disconnected')
        self.transitions.add_transition(trigger='disconnect', source='disconnected', dest='disconnected')

    def __hash__(self):
        return hash(self.client_id)

    def __eq__(self, other):
        other = getattr(other,'client_id', other)
        return self.client_id == other

    async def start(self, broker=None):
        if broker is not None:
            self._broker = broker
            if self._delivery_task is not None:
                raise RuntimeError("Already running")
            await broker._tg.spawn(self._delivery_loop)

    async def stop(self):
        if self._delivery_task is not None:
            await self._delivery_task.cancel()
            await self._delivery_stopped.wait()
        self._broker = None  # break ref loop

    async def put_message(self, app_message):
        if not app_message.topic:
            self.logger.warning("[MQTT-4.7.3-1] - %s invalid TOPIC sent in PUBLISH message,closing connection", self.client_id)
            raise MQTTException("[MQTT-4.7.3-1] - %s invalid TOPIC sent in PUBLISH message,closing connection" % self.client_id)
        if "#" in app_message.topic or "+" in app_message.topic:
            self.logger.warning("[MQTT-3.3.2-2] - %s invalid TOPIC sent in PUBLISH message, closing connection", self.client_id)
            raise MQTTException("[MQTT-3.3.2-2] - %s invalid TOPIC sent in PUBLISH message, closing connection" % self.client_id)
        if app_message.qos == QOS_0 and self._delivered_message_queue.qsize() >= 9999:
            self.logger.warning("delivered messages queue full. QOS_0 message discarded")
        else:
            await self._delivered_message_queue.put(app_message)

    async def get_next_message(self):
        """Client: get the next message"""
        m = await self._delivered_message_queue.get()
        # split up so that a breakpoint may be set
        return m

    async def _delivery_loop(self):
        """Server: process incoming messages"""
        try:
            async with anyio.open_cancel_scope() as scope:
                self._delivery_task = scope
                broker = self._broker
                broker.logger.debug("%s handling message delivery", self.client_id)
                while True:
                    app_message = await self.get_next_message()
                    await self._plugins_manager.fire_event(EVENT_BROKER_MESSAGE_RECEIVED,
                                                client_id=self.client_id,      
                                                message=app_message)

                    await broker.broadcast_message(self, app_message.topic, app_message.data, qos=app_message.qos, retain=app_message.publish_packet.retain_flag)
        finally:
            broker.logger.debug("%s finished message delivery", self.client_id)
            self._delivery_task = None
            await self._delivery_stopped.set()

    @property
    def next_packet_id(self):
        self._packet_id += 1
        if self._packet_id > 65535:
            self._packet_id = 1
        limit = self._packet_id
        while self._packet_id in self.inflight_in or self._packet_id in self.inflight_out:
            self._packet_id += 1
            if self._packet_id > 65535:
                self._packet_id = 1
            if self._packet_id == limit:
                raise DistMQTTException("More than 65535 messages pending. No free packet ID")

        return self._packet_id

    @property
    def inflight_in_count(self):
        return len(self.inflight_in)

    @property
    def inflight_out_count(self):
        return len(self.inflight_out)

    @property
    def retained_messages_count(self):
        return self.retained_messages.qsize()

    def __repr__(self):
        return type(self).__name__ + '(clientId={0}, state={1})'.format(self.client_id, self.transitions.state)

    def __getstate__(self):
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        # del state['transitions']
        del state['retained_messages']
        del state['_delivered_message_queue']
        del state['_delivery_task']
        del state['_delivery_stopped']
        del state['_broker']
        return state

    def __setstate(self, state):
        self.__dict__.update(state)
        self.retained_messages = anyio.create_queue(9999)
        self._delivered_message_queue = anyio.create_queue(9999)

    def __eq__(self, other):
        return self.client_id == other.client_id



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.

from distmqtt.mqtt.packet import MQTTPacket, MQTTFixedHeader, SUBACK, PacketIdVariableHeader, MQTTPayload, MQTTVariableHeader
from distmqtt.errors import DistMQTTException, NoDataException
from distmqtt.adapters import StreamAdapter
from distmqtt.codecs import bytes_to_int, int_to_bytes, read_or_raise


class SubackPayload(MQTTPayload):

    __slots__ = ('return_codes',)

    RETURN_CODE_00 = 0x00
    RETURN_CODE_01 = 0x01
    RETURN_CODE_02 = 0x02
    RETURN_CODE_80 = 0x80

    def __init__(self, return_codes=[]):
        super().__init__()
        self.return_codes = return_codes

    def __repr__(self):
        return type(self).__name__ + '(return_codes={0})'.format(repr(self.return_codes))

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        out = b''
        for return_code in self.return_codes:
            out += int_to_bytes(return_code, 1)
        return out

    @classmethod
    async def from_stream(cls, reader: StreamAdapter, fixed_header: MQTTFixedHeader,
                    variable_header: MQTTVariableHeader):
        return_codes = []
        bytes_to_read = fixed_header.remaining_length - variable_header.bytes_length
        for i in range(0, bytes_to_read):
            try:
                return_code_byte = await read_or_raise(reader, 1)
                return_code = bytes_to_int(return_code_byte)
                return_codes.append(return_code)
            except NoDataException:
                break
        return cls(return_codes)


class SubackPacket(MQTTPacket):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = SubackPayload

    def __init__(self, fixed: MQTTFixedHeader=None, variable_header: PacketIdVariableHeader=None, payload=None):
        if fixed is None:
            header = MQTTFixedHeader(SUBACK, 0x00)
        else:
            if fixed.packet_type is not SUBACK:
                raise DistMQTTException("Invalid fixed packet type %s for SubackPacket init" % fixed.packet_type)
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, packet_id, return_codes):
        variable_header = cls.VARIABLE_HEADER(packet_id)
        payload = cls.PAYLOAD(return_codes)
        return cls(variable_header=variable_header, payload=payload)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import unittest
import anyio

from distmqtt.codecs import (
    bytes_to_hex_str,
    bytes_to_int,
    decode_string,
    encode_string,
)
from distmqtt.adapters import BufferAdapter


class TestCodecs(unittest.TestCase):
    def test_bytes_to_hex_str(self):
        ret = bytes_to_hex_str(b'\x7f')
        self.assertEqual(ret, '0x7f')

    def test_bytes_to_int(self):
        ret = bytes_to_int(b'\x7f')
        self.assertEqual(ret, 127)
        ret = bytes_to_int(b'\xff\xff')
        self.assertEqual(ret, 65535)

    def test_decode_string(self):
        async def test_coro():
            stream = BufferAdapter(b'\x00\x02AA')
            ret = await decode_string(stream)
            self.assertEqual(ret, 'AA')
        anyio.run(test_coro)

    def test_encode_string(self):
        encoded = encode_string('AA')
        self.assertEqual(b'\x00\x02AA', encoded)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
from datetime import datetime
from distmqtt.mqtt.packet import PUBLISH
from distmqtt.codecs import int_to_bytes_str
import anyio
import sys
from collections import deque


DOLLAR_SYS_ROOT = '$SYS/broker/'
STAT_BYTES_SENT = 'bytes_sent'
STAT_BYTES_RECEIVED = 'bytes_received'
STAT_MSG_SENT = 'messages_sent'
STAT_MSG_RECEIVED = 'messages_received'
STAT_PUBLISH_SENT = 'publish_sent'
STAT_PUBLISH_RECEIVED = 'publish_received'
STAT_START_TIME = 'start_time'
STAT_CLIENTS_MAXIMUM = 'clients_maximum'
STAT_CLIENTS_CONNECTED = 'clients_connected'
STAT_CLIENTS_DISCONNECTED = 'clients_disconnected'


class BrokerSysPlugin:
    def __init__(self, context):
        self.context = context
        # Broker statistics initialization
        self._stats = dict()
        self._sys_handle = None

    def _clear_stats(self):
        """
        Initializes broker statistics data structures
        """
        for stat in (STAT_BYTES_RECEIVED,
                     STAT_BYTES_SENT,
                     STAT_MSG_RECEIVED,
                     STAT_MSG_SENT,
                     STAT_CLIENTS_MAXIMUM,
                     STAT_CLIENTS_CONNECTED,
                     STAT_CLIENTS_DISCONNECTED,
                     STAT_PUBLISH_RECEIVED,
                     STAT_PUBLISH_SENT):
            self._stats[stat] = 0

    async def _broadcast_sys_topic(self, topic_basename, data):
        return await self.context.broadcast_message(topic_basename, data)

    async def on_broker_pre_start(self, *args, **kwargs):
        self._clear_stats()

    async def on_broker_post_start(self, *args, **kwargs):
        self._stats[STAT_START_TIME] = datetime.now()
        from distmqtt.version import get_version
        version = 'DistMQTT version ' + get_version()
        await self.context.broadcast_message(DOLLAR_SYS_ROOT + 'version', version.encode(), retain=True)

        # Start $SYS topics management
        try:
            sys_interval = int(self.context.config.get('sys_interval', 0))
            if sys_interval > 0:
                self.context.logger.debug("Setup $SYS broadcasting every %d secondes", sys_interval)
                evt = anyio.create_event()
                await self.context._broker_instance._tg.spawn(self.broadcast_dollar_sys_topics_loop, sys_interval, evt)
                await evt.wait()
            else:
                self.context.logger.debug("$SYS disabled")
        except KeyError:
            pass
            # 'sys_internal' config parameter not found

    async def on_broker_pre_stop(self, *args, **kwargs):
        # Stop $SYS topics broadcasting
        if self.sys_handle:
            await self.sys_handle.cancel()

    async def broadcast_dollar_sys_topics_loop(self, interval, evt):
        async with anyio.open_cancel_scope() as scope:
            self.sys_handle = scope
            await evt.set()
            while True:
                await anyio.sleep(interval)
                await self.broadcast_dollar_sys_topics()
    async def broadcast_dollar_sys_topics(self):
        """
        Broadcast dynamic $SYS topics updates and reschedule next execution depending on 'sys_interval' config
        parameter.
        """

        # Update stats
        uptime = datetime.now() - self._stats[STAT_START_TIME]
        client_connected = self._stats[STAT_CLIENTS_CONNECTED]
        client_disconnected = self._stats[STAT_CLIENTS_DISCONNECTED]
        inflight_in = 0
        inflight_out = 0
        messages_stored = 0
        for session in self.context.sessions:
            inflight_in += session.inflight_in_count
            inflight_out += session.inflight_out_count
            messages_stored += session.retained_messages_count
        messages_stored += len(self.context.retained_messages)
        subscriptions_count = 0
        for topic in self.context.subscriptions:
            subscriptions_count += len(self.context.subscriptions[topic])

        # Broadcast updates
        await self._broadcast_sys_topic('load/bytes/received', int_to_bytes_str(self._stats[STAT_BYTES_RECEIVED]))
        await self._broadcast_sys_topic('load/bytes/sent', int_to_bytes_str(self._stats[STAT_BYTES_SENT]))
        await self._broadcast_sys_topic('messages/received', int_to_bytes_str(self._stats[STAT_MSG_RECEIVED]))
        await self._broadcast_sys_topic('messages/sent', int_to_bytes_str(self._stats[STAT_MSG_SENT]))
        await self._broadcast_sys_topic('time', str(datetime.now()).encode('utf-8'))
        await self._broadcast_sys_topic('uptime', int_to_bytes_str(int(uptime.total_seconds())))
        await self._broadcast_sys_topic('uptime/formated', str(uptime).encode('utf-8'))
        await self._broadcast_sys_topic('clients/connected', int_to_bytes_str(client_connected))
        await self._broadcast_sys_topic('clients/disconnected', int_to_bytes_str(client_disconnected))
        await self._broadcast_sys_topic('clients/maximum', int_to_bytes_str(self._stats[STAT_CLIENTS_MAXIMUM]))
        await self._broadcast_sys_topic('clients/total', int_to_bytes_str(client_connected + client_disconnected))
        await self._broadcast_sys_topic('messages/inflight', int_to_bytes_str(inflight_in + inflight_out))
        await self._broadcast_sys_topic('messages/inflight/in', int_to_bytes_str(inflight_in))
        await self._broadcast_sys_topic('messages/inflight/out', int_to_bytes_str(inflight_out))
        await self._broadcast_sys_topic('messages/inflight/stored', int_to_bytes_str(messages_stored))
        await self._broadcast_sys_topic('messages/publish/received', int_to_bytes_str(self._stats[STAT_PUBLISH_RECEIVED]))
        await self._broadcast_sys_topic('messages/publish/sent', int_to_bytes_str(self._stats[STAT_PUBLISH_SENT]))
        await self._broadcast_sys_topic('messages/retained/count', int_to_bytes_str(len(self.context.retained_messages)))
        await self._broadcast_sys_topic('messages/subscriptions/count', int_to_bytes_str(subscriptions_count))

    async def on_mqtt_packet_received(self, *args, **kwargs):
        packet = kwargs.get('packet')
        if packet:
            packet_size = packet.bytes_length
            self._stats[STAT_BYTES_RECEIVED] += packet_size
            self._stats[STAT_MSG_RECEIVED] += 1
            if packet.fixed_header.packet_type == PUBLISH:
                self._stats[STAT_PUBLISH_RECEIVED] += 1

    async def on_mqtt_packet_sent(self, *args, **kwargs):
        packet = kwargs.get('packet')
        if packet:
            packet_size = packet.bytes_length
            self._stats[STAT_BYTES_SENT] += packet_size
            self._stats[STAT_MSG_SENT] += 1
            if packet.fixed_header.packet_type == PUBLISH:
                self._stats[STAT_PUBLISH_SENT] += 1

    async def on_broker_client_connected(self, *args, **kwargs):
        self._stats[STAT_CLIENTS_CONNECTED] += 1
        self._stats[STAT_CLIENTS_MAXIMUM] = max(self._stats[STAT_CLIENTS_MAXIMUM], self._stats[STAT_CLIENTS_CONNECTED])

    async def on_broker_client_disconnected(self, *args, **kwargs):
        self._stats[STAT_CLIENTS_CONNECTED] -= 1
        self._stats[STAT_CLIENTS_DISCONNECTED] += 1



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import anyio

from distmqtt.codecs import bytes_to_hex_str, decode_packet_id, int_to_bytes, read_or_raise
from distmqtt.errors import CodecException, MQTTException, NoDataException
from distmqtt.adapters import StreamAdapter
from datetime import datetime
from struct import unpack


RESERVED_0 = 0x00
CONNECT = 0x01
CONNACK = 0x02
PUBLISH = 0x03
PUBACK = 0x04
PUBREC = 0x05
PUBREL = 0x06
PUBCOMP = 0x07
SUBSCRIBE = 0x08
SUBACK = 0x09
UNSUBSCRIBE = 0x0a
UNSUBACK = 0x0b
PINGREQ = 0x0c
PINGRESP = 0x0d
DISCONNECT = 0x0e
RESERVED_15 = 0x0f


class MQTTFixedHeader:

    __slots__ = ('packet_type', 'remaining_length', 'flags')

    def __init__(self, packet_type, flags=0, length=0):
        self.packet_type = packet_type
        self.remaining_length = length
        self.flags = flags

    def to_bytes(self):
        def encode_remaining_length(length: int):
            encoded = bytearray()
            while True:
                length_byte = length % 0x80
                length //= 0x80
                if length > 0:
                    length_byte |= 0x80
                encoded.append(length_byte)
                if length <= 0:
                    break
            return encoded

        out = bytearray()
        packet_type = 0
        try:
            packet_type = (self.packet_type << 4) | self.flags
            out.append(packet_type)
        except OverflowError:
            raise CodecException('packet_type encoding exceed 1 byte length: value=%d', packet_type)

        encoded_length = encode_remaining_length(self.remaining_length)
        out.extend(encoded_length)

        return out

    async def to_stream(self, writer: StreamAdapter):
        await writer.write(self.to_bytes())

    @property
    def bytes_length(self):
        return len(self.to_bytes())

    @classmethod
    async def from_stream(cls, reader: StreamAdapter):
        """
        Read and decode MQTT message fixed header from stream
        :return: FixedHeader instance
        """
        async def decode_remaining_length():
            """
            Decode message length according to MQTT specifications
            :return:
            """
            multiplier = 1
            value = 0
            buffer = bytearray()
            while True:
                encoded_byte = await reader.read(1)
                int_byte = unpack('!B', encoded_byte)
                buffer.append(int_byte[0])
                value += (int_byte[0] & 0x7f) * multiplier
                if (int_byte[0] & 0x80) == 0:
                    break
                else:
                    multiplier *= 128
                    if multiplier > 128 * 128 * 128:
                        raise MQTTException("Invalid remaining length bytes:%s, packet_type=%d" % (bytes_to_hex_str(buffer), msg_type))
            return value

        try:
            byte1 = await read_or_raise(reader, 1)
            int1 = unpack('!B', byte1)
            msg_type = (int1[0] & 0xf0) >> 4
            flags = int1[0] & 0x0f
            remain_length = await decode_remaining_length()

            return cls(msg_type, flags, remain_length)
        except NoDataException:
            return None

    def __repr__(self):
        return type(self).__name__ + '(length={0}, flags={1})'.\
            format(self.remaining_length, hex(self.flags))


class MQTTVariableHeader:
    def __init__(self):
        pass

    async def to_stream(self, writer: anyio.abc.Stream):
        await writer.write(self.to_bytes())

    def to_bytes(self) -> bytes:
        """
        Serialize header data to a byte array conforming to MQTT protocol
        :return: serialized data
        """

    @property
    def bytes_length(self):
        return len(self.to_bytes())

    @classmethod
    async def from_stream(cls, reader: StreamAdapter, fixed_header: MQTTFixedHeader):
        pass


class PacketIdVariableHeader(MQTTVariableHeader):

    __slots__ = ('packet_id',)

    def __init__(self, packet_id):
        super().__init__()
        self.packet_id = packet_id

    def to_bytes(self):
        out = b''
        out += int_to_bytes(self.packet_id, 2)
        return out

    @classmethod
    async def from_stream(cls, reader: StreamAdapter, fixed_header: MQTTFixedHeader):
        packet_id = await decode_packet_id(reader)
        return cls(packet_id)

    def __repr__(self):
        return type(self).__name__ + '(packet_id={0})'.format(self.packet_id)


class MQTTPayload:
    def __init__(self):
        pass

    async def to_stream(self, writer: anyio.abc.Stream):
        await writer.write(self.to_bytes())

    def to_bytes(self, fixed_header: MQTTFixedHeader, variable_header: MQTTVariableHeader):
        pass

    @classmethod
    async def from_stream(cls, reader: anyio.abc.Stream, fixed_header: MQTTFixedHeader,
                    variable_header: MQTTVariableHeader):
        pass


class MQTTPacket:

    __slots__ = ('fixed_header', 'variable_header', 'payload', 'protocol_ts')

    FIXED_HEADER = MQTTFixedHeader
    VARIABLE_HEADER = None
    PAYLOAD = None

    def __init__(self, fixed: MQTTFixedHeader, variable_header: MQTTVariableHeader=None, payload: MQTTPayload=None):
        self.fixed_header = fixed
        self.variable_header = variable_header
        self.payload = payload
        self.protocol_ts = None

    async def to_stream(self, writer: anyio.abc.Stream):
        await writer.write(self.to_bytes())
        self.protocol_ts = datetime.now()

    def to_bytes(self) -> bytes:
        if self.variable_header:
            variable_header_bytes = self.variable_header.to_bytes()
        else:
            variable_header_bytes = b''
        if self.payload:
            payload_bytes = self.payload.to_bytes(self.fixed_header, self.variable_header)
        else:
            payload_bytes = b''

        self.fixed_header.remaining_length = len(variable_header_bytes) + len(payload_bytes)
        fixed_header_bytes = self.fixed_header.to_bytes()

        return fixed_header_bytes + variable_header_bytes + payload_bytes

    @classmethod
    async def from_stream(cls, reader: StreamAdapter, fixed_header=None, variable_header=None):
        if fixed_header is None:
            fixed_header = await cls.FIXED_HEADER.from_stream(reader)
        if cls.VARIABLE_HEADER:
            if variable_header is None:
                variable_header = await cls.VARIABLE_HEADER.from_stream(reader, fixed_header)
        else:
            variable_header = None
        if cls.PAYLOAD:
            payload = await cls.PAYLOAD.from_stream(reader, fixed_header, variable_header)
        else:
            payload = None

        if fixed_header and not variable_header and not payload:
            instance = cls(fixed_header)
        elif fixed_header and not payload:
            instance = cls(fixed_header, variable_header)
        else:
            instance = cls(fixed_header, variable_header, payload)
        instance.protocol_ts = datetime.now()
        return instance

    @property
    def bytes_length(self):
        return len(self.to_bytes())

    def __repr__(self):
        return type(self).__name__ + '(ts={0!s}, fixed={1!r}, variable={2!r}, payload={3!r})'.\
            format(self.protocol_ts, self.fixed_header, self.variable_header, self.payload)



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.
import logging
import collections
import itertools
import anyio

from distmqtt.mqtt import packet_class
from distmqtt.mqtt.connack import ConnackPacket
from distmqtt.mqtt.connect import ConnectPacket
from distmqtt.mqtt.packet import (
    RESERVED_0, CONNECT, CONNACK, PUBLISH, PUBACK, PUBREC, PUBREL, PUBCOMP,
    SUBSCRIBE, SUBACK, UNSUBSCRIBE, UNSUBACK, PINGREQ, PINGRESP, DISCONNECT,
    RESERVED_15, MQTTFixedHeader)
from distmqtt.mqtt.pingresp import PingRespPacket
from distmqtt.mqtt.pingreq import PingReqPacket
from distmqtt.mqtt.publish import PublishPacket
from distmqtt.mqtt.pubrel import PubrelPacket
from distmqtt.mqtt.puback import PubackPacket
from distmqtt.mqtt.pubrec import PubrecPacket
from distmqtt.mqtt.pubcomp import PubcompPacket
from distmqtt.mqtt.suback import SubackPacket
from distmqtt.mqtt.subscribe import SubscribePacket
from distmqtt.mqtt.unsubscribe import UnsubscribePacket
from distmqtt.mqtt.unsuback import UnsubackPacket
from distmqtt.mqtt.disconnect import DisconnectPacket
from distmqtt.adapters import StreamAdapter
from distmqtt.session import Session, OutgoingApplicationMessage, IncomingApplicationMessage, INCOMING, OUTGOING
from distmqtt.mqtt.constants import QOS_0, QOS_1, QOS_2
from distmqtt.plugins.manager import PluginManager
from distmqtt.errors import DistMQTTException, MQTTException, NoDataException, InvalidStateError
from distmqtt.utils import Future, CancelledError


EVENT_MQTT_PACKET_SENT = 'mqtt_packet_sent'
EVENT_MQTT_PACKET_RECEIVED = 'mqtt_packet_received'

PACKET_TYPES = {
    CONNACK: ("connack", False),
    SUBSCRIBE: ("subscribe", True),
    UNSUBSCRIBE: ("unsubscribe", True),
    SUBACK: ("suback", False),
    UNSUBACK: ("unsuback", False),
    PUBACK: ("puback", False),
    PUBREC: ("pubrec", False),
    PUBREL: ("pubrel", False),
    PUBCOMP: ("pubcomp", False),
    PINGREQ: ("pingreq", False),
    PINGRESP: ("pingresp", False),
    PUBLISH: ("publish", True),
    DISCONNECT: ("disconnect", True),
}

class ProtocolHandlerException(Exception):
    pass


class ProtocolHandler:
    """
    Class implementing the MQTT communication protocol using async features
    """

    def __init__(self, plugins_manager: PluginManager, session: Session=None):
        self.logger = logging.getLogger(__name__+"."+self.__class__.__name__)
        if session:
            self._init_session(session)
        else:
            self.session = None
        self.stream = None
        self.plugins_manager = plugins_manager
        self._tg = plugins_manager._tg

        self._reader_task = None
        self._sender_task = None
        self._reader_stopped = anyio.create_event()
        self._sender_stopped = anyio.create_event()
        self._send_q = anyio.create_queue(10)

        self._puback_waiters = dict()
        self._pubrec_waiters = dict()
        self._pubrel_waiters = dict()
        self._pubcomp_waiters = dict()

        self._disconnecting = False
        self._disconnect_waiter = None
        self._write_lock = anyio.create_lock()

    def _init_session(self, session: Session):
        assert session
        log = logging.getLogger(__name__)
        self.session = session
        self.logger = logging.LoggerAdapter(log, {'client_id': self.session.client_id})
        self.keepalive_timeout = self.session.keep_alive
        if self.keepalive_timeout <= 0:
            self.keepalive_timeout = None

    async def attach(self, session, stream: StreamAdapter):
        if self.session:
            raise ProtocolHandlerException("Handler is already attached to a session")
        self._init_session(session)
        self.stream = stream
        await self._tg.spawn(self._sender_loop)

    async def detach(self):
        self.session = None
        self.stream = None

    def _is_attached(self):
        if self.session:
            return True
        else:
            return False

    async def start(self):
        self._disconnect_waiter = anyio.create_event()
        if not self._is_attached():
            raise ProtocolHandlerException("Handler is not attached to a stream")
        await self._tg.spawn(self._reader_loop)

        self.logger.debug("Handler tasks started")
        await self._retry_deliveries()
        self.logger.debug("%s %s ready",
                'Broker' if 'Broker' in type(self).__name__ else 'Client',
                self.session.client_id if self.session else '?')

    async def stop(self):
        # Stop messages flow waiter
        self.logger.debug("%s %s stopping",
                'Broker' if 'Broker' in type(self).__name__ else 'Client',
                self.session.client_id if self.session else '?')
        await self._stop_waiters()
        t, self._reader_task = self._reader_task, None
        if t:
            self.logger.debug("waiting for reader %s to be stopped", t)
            await t.cancel_scope.cancel()
            await self._reader_stopped.wait()
        t, self._sender_task = self._sender_task, None
        if t:
            self.logger.debug("waiting for sender %s to be stopped", t)
            await self._send_q.put(None)
            await self._sender_stopped.wait()
        self.logger.debug("closing writer")
        if self.stream is not None:
            await self.stream.close()
        if self._disconnect_waiter is not None:
            await self._disconnect_waiter.set()

    async def _stop_waiters(self):
        if len(self._puback_waiters):
            self.logger.debug("Stopping %d puback waiters", len(self._puback_waiters))
        if len(self._pubcomp_waiters):
            self.logger.debug("Stopping %d pucomp waiters", len(self._pubcomp_waiters))
        if len(self._pubrec_waiters):
            self.logger.debug("Stopping %d purec waiters", len(self._pubrec_waiters))
        if len(self._pubrel_waiters):
            self.logger.debug("Stopping %d purel waiters", len(self._pubrel_waiters))
        for waiter in itertools.chain(
                self._puback_waiters.values(),
                self._pubcomp_waiters.values(),
                self._pubrec_waiters.values(),
                self._pubrel_waiters.values()):
            await waiter.cancel()

    async def _retry_deliveries(self):
        """
        Handle [MQTT-4.4.0-1] by resending PUBLISH and PUBREL messages for pending out messages
        :return:
        """
        done = pending = 0
        self.logger.debug("Begin messages delivery retries")

        async def process_one(message):
            async with anyio.move_on_after(10):
                try:
                    await self._handle_message_flow(message)
                except CancelledError:
                    pass

                nonlocal done
                done += 1

        async with anyio.create_task_group() as tg:
            for message in itertools.chain(self.session.inflight_in.values(), self.session.inflight_out.values()):
                pending += 1
                await tg.spawn(process_one, message)
        pending -= done

        self.logger.debug("%d messages redelivered", done)
        self.logger.debug("%d messages not redelivered due to timeout", pending)
        self.logger.debug("End messages delivery retries")

    async def mqtt_publish(self, topic, data, qos, retain):
        """
        Sends a MQTT publish message and manages messages flows.
        This methods doesn't return until the message has been acknowledged by receiver or timeout occur
        :param topic: MQTT topic to publish
        :param data:  data to send on topic
        :param qos: quality of service to use for message flow. Can be QOS_0, QOS_1 or QOS_2
        :param retain: retain message flag
        :param ack_timeout: acknowledge timeout. If set, this method will return a TimeOut error if the acknowledgment
        is not completed before ack_timeout second
        :return: ApplicationMessage used during inflight operations
        """
        if qos in (QOS_1, QOS_2):
            packet_id = self.session.next_packet_id
            if packet_id in self.session.inflight_out:
                raise DistMQTTException("A message with the same packet ID '%d' is already in flight" % packet_id)
        else:
            packet_id = None

        message = OutgoingApplicationMessage(packet_id, topic, qos, data, retain)
        await self._handle_message_flow(message)

        return message

    async def _handle_message_flow(self, app_message):
        """
        Handle protocol flow for incoming and outgoing messages, depending on service level and according to MQTT
        spec. paragraph 4.3-Quality of Service levels and protocol flows
        :param app_message: PublishMessage to handle
        :return: nothing.
        """
        try:
            if app_message.qos == QOS_0:
                await self._handle_qos0_message_flow(app_message)
            elif app_message.qos == QOS_1:
                await self._handle_qos1_message_flow(app_message)
            elif app_message.qos == QOS_2:
                await self._handle_qos2_message_flow(app_message)
            else:
                raise DistMQTTException("Unexcepted QOS value '%d" % str(app_message.qos))
        except CancelledError:
            pass

    async def _handle_qos0_message_flow(self, app_message):
        """
        Handle QOS_0 application message acknowledgment
        For incoming messages, this method stores the message
        For outgoing messages, this methods sends PUBLISH
        :param app_message:
        :return:
        """
        assert app_message.qos == QOS_0
        if app_message.direction == OUTGOING:
            packet = app_message.build_publish_packet()
            # Send PUBLISH packet
            await self._send_packet(packet)
            app_message.publish_packet = packet
        elif app_message.direction == INCOMING:
            if app_message.publish_packet.dup_flag:
                self.logger.warning("[MQTT-3.3.1-2] DUP flag must set to 0 for QOS 0 message. Message ignored: %r",
                                    app_message.publish_packet)
            else:
                await self.session.put_message(app_message)

    async def _handle_qos1_message_flow(self, app_message):
        """
        Handle QOS_1 application message acknowledgment
        For incoming messages, this method stores the message and reply with PUBACK
        For outgoing messages, this methods sends PUBLISH and waits for the corresponding PUBACK
        :param app_message:
        :return:
        """
        assert app_message.qos == QOS_1
        if app_message.puback_packet:
            raise DistMQTTException("Message '%d' has already been acknowledged" % app_message.packet_id)
        if app_message.direction == OUTGOING:
            if app_message.packet_id not in self.session.inflight_out:
                # Store message in session
                self.session.inflight_out[app_message.packet_id] = app_message
            if app_message.publish_packet is not None:
                # A Publish packet has already been sent, this is a retry
                publish_packet = app_message.build_publish_packet(dup=True)
            else:
                publish_packet = app_message.build_publish_packet()

            # Wait for puback
            waiter = Future()
            self._puback_waiters[app_message.packet_id] = waiter
            # Send PUBLISH packet
            try:
                await self._send_packet(publish_packet)
                app_message.publish_packet = publish_packet
                app_message.puback_packet = await waiter.get()
            finally:
                del self._puback_waiters[app_message.packet_id]

            # Discard inflight message
            del self.session.inflight_out[app_message.packet_id]
        elif app_message.direction == INCOMING:
            # Initiate delivery
            self.logger.debug("Add message to delivery")
            await self.session.put_message(app_message)
            # Send PUBACK
            puback = PubackPacket.build(app_message.packet_id)
            await self._send_packet(puback)
            app_message.puback_packet = puback

    async def _handle_qos2_message_flow(self, app_message):
        """
        Handle QOS_2 application message acknowledgment
        For incoming messages, this method stores the message, sends PUBREC, waits for PUBREL, initiate delivery
        and send PUBCOMP
        For outgoing messages, this methods sends PUBLISH, waits for PUBREC, discards messages and wait for PUBCOMP
        :param app_message:
        :return:
        """
        assert app_message.qos == QOS_2
        if app_message.direction == OUTGOING:
            if app_message.pubrel_packet and app_message.pubcomp_packet:
                raise DistMQTTException("Message '%d' has already been acknowledged" % app_message.packet_id)
            if not app_message.pubrel_packet:
                # Store message
                if app_message.publish_packet is not None:
                    # This is a retry flow, no need to store just check the message exists in session
                    if app_message.packet_id not in self.session.inflight_out:
                        raise DistMQTTException("Unknown inflight message '%d' in session" % app_message.packet_id)
                    publish_packet = app_message.build_publish_packet(dup=True)
                else:
                    # Store message in session
                    self.session.inflight_out[app_message.packet_id] = app_message
                    publish_packet = app_message.build_publish_packet()

                # Wait PUBREC
                if app_message.packet_id in self._pubrec_waiters:
                    # PUBREC waiter already exists for this packet ID
                    message = "Can't add PUBREC waiter, a waiter already exists for message Id '%s'" \
                              % app_message.packet_id
                    self.logger.warning(message)
                    raise DistMQTTException(message)
                waiter = Future()
                self._pubrec_waiters[app_message.packet_id] = waiter
                # Send PUBLISH packet
                try:
                    await self._send_packet(publish_packet)
                    app_message.publish_packet = publish_packet
                    app_message.pubrec_packet = await waiter.get()
                finally:
                    del self._pubrec_waiters[app_message.packet_id]

            if not app_message.pubcomp_packet:
                # Wait for PUBCOMP
                waiter = Future()
                self._pubcomp_waiters[app_message.packet_id] = waiter
                # Send pubrel
                try:
                    app_message.pubrel_packet = PubrelPacket.build(app_message.packet_id)
                    await self._send_packet(app_message.pubrel_packet)
                    app_message.pubcomp_packet = await waiter.get()
                finally:
                    del self._pubcomp_waiters[app_message.packet_id]
            # Discard inflight message
            del self.session.inflight_out[app_message.packet_id]
        elif app_message.direction == INCOMING:
            self.session.inflight_in[app_message.packet_id] = app_message
            # Wait PUBREL
            if app_message.packet_id in self._pubrel_waiters and not self._pubrel_waiters[app_message.packet_id].done():
                # PUBREL waiter already exists for this packet ID
                message = "A waiter already exists for message Id '%s', canceling it" \
                          % app_message.packet_id
                self.logger.warning(message)
                await self._pubrel_waiters[app_message.packet_id].cancel()
            waiter = Future()
            self._pubrel_waiters[app_message.packet_id] = waiter
            # Send pubrec
            try:
                pubrec_packet = PubrecPacket.build(app_message.packet_id)
                await self._send_packet(pubrec_packet)
                app_message.pubrec_packet = pubrec_packet
                app_message.pubrel_packet = await waiter.get()
            finally:
                if self._pubrel_waiters.get(app_message.packet_id) is waiter:
                    del self._pubrel_waiters[app_message.packet_id]
            # Initiate delivery and discard message
            await self.session.put_message(app_message)
            del self.session.inflight_in[app_message.packet_id]
            # Send pubcomp
            pubcomp_packet = PubcompPacket.build(app_message.packet_id)
            await self._send_packet(pubcomp_packet)
            app_message.pubcomp_packet = pubcomp_packet

    async def _reader_loop(self):
        self.logger.debug("%s Starting reader coro", self.session.client_id)
        keepalive_timeout = self.session.keep_alive
        if keepalive_timeout <= 0:
            keepalive_timeout = None

        try:
            async with anyio.create_task_group() as tg:
                self._reader_task = tg
                while True:
                    try:
                        async with anyio.fail_after(keepalive_timeout):
                            fixed_header = await MQTTFixedHeader.from_stream(self.stream)
                        if fixed_header is None:
                            self.logger.debug("%s No more data (EOF received), stopping reader coro", self.session.client_id)
                            break

                        if fixed_header.packet_type == RESERVED_0 or fixed_header.packet_type == RESERVED_15:
                            self.logger.warning("%s Received reserved packet, which is forbidden: closing connection",
                                                self.session.client_id)
                            await self.handle_connection_closed()
                            break
                        cls = packet_class(fixed_header)
                        packet = await cls.from_stream(self.stream, fixed_header=fixed_header)
                        self.logger.debug("< %s %r",'B' if 'Broker' in type(self).__name__ else 'C', packet)
                        await self.plugins_manager.fire_event(
                            EVENT_MQTT_PACKET_RECEIVED, packet=packet, session=self.session)
                        try:
                            pt,direct = PACKET_TYPES[packet.fixed_header.packet_type]
                            fn = getattr(self,'handle_'+pt)
                        except (KeyError,AttributeError):
                            self.logger.warning("%s Unhandled packet type: %s",
                                            self.session.client_id, packet.fixed_header.packet_type)
                        else:
                            try:
                                if direct:
                                    await fn(packet)
                                else:
                                    await tg.spawn(fn, packet)
                            except StopAsyncIteration:
                                break

                    except MQTTException:
                        self.logger.debug("Message discarded")
                    except TimeoutError:
                        self.logger.debug("%s Input stream read timeout", self.session.client_id if self.session else '?')
                        await self.handle_read_timeout()
                    except NoDataException:
                        self.logger.debug("%s No data available", self.session.client_id)
                        break # XXX
                    except anyio.get_cancelled_exc_class():
                        self.logger.warning("%s CANCEL", type(self).__name__)
                        raise
                    except BaseException as e:
                        self.logger.warning("%s Unhandled exception in reader coro", type(self).__name__, exc_info=e)
                        raise
                await tg.cancel_scope.cancel()
        finally:
            async with anyio.open_cancel_scope(shield=True):
                self.logger.debug("%s %s coro stopped",
                        'Broker' if 'Broker' in type(self).__name__ else 'Client',
                        self.session.client_id if self.session else '?')
                await self._reader_stopped.set()
                if self._reader_task is not None:
                    self._reader_task = None
                    await self.handle_connection_closed()

    async def _send_packet(self, packet):
        await self._send_q.put(packet)

    async def _sender_loop(self):
        keepalive_timeout = self.session.keep_alive
        if keepalive_timeout <= 0:
            keepalive_timeout = None

        try:
            async with anyio.open_cancel_scope() as scope:
                self._sender_task = scope
                while True:
                    packet = None
                    async with anyio.move_on_after(keepalive_timeout):
                        packet = await self._send_q.get()
                        if packet is None:  # closing
                            break
                    if packet is None:  # timeout
                        await self.handle_write_timeout()
                        continue
                    self.logger.debug("%s > %r",'B' if 'Broker' in type(self).__name__ else 'C', packet)
                    await packet.to_stream(self.stream)
                    await self.plugins_manager.fire_event(EVENT_MQTT_PACKET_SENT, packet=packet, session=self.session)
        except ConnectionResetError as cre:
            await self.handle_connection_closed()
        except anyio.get_cancelled_exc_class():
            raise
        except BaseException as e:
            self.logger.warning("Unhandled exception", exc_info=e)
            raise
        finally:
            async with anyio.open_cancel_scope(shield=True):
                await self._sender_stopped.set()
            self._sender_task = None

    async def handle_write_timeout(self):
        self.logger.debug('%s write timeout unhandled' % self.session.client_id)

    async def handle_read_timeout(self):
        self.logger.debug('%s read timeout unhandled' % self.session.client_id)

    async def handle_connack(self, connack: ConnackPacket):
        self.logger.debug('%s CONNACK unhandled' % self.session.client_id)

    async def handle_connect(self, connect: ConnectPacket):
        self.logger.debug('%s CONNECT unhandled' % self.session.client_id)

    async def handle_subscribe(self, subscribe: SubscribePacket):
        self.logger.debug('%s SUBSCRIBE unhandled' % self.session.client_id)

    async def handle_unsubscribe(self, subscribe: UnsubscribePacket):
        self.logger.debug('%s UNSUBSCRIBE unhandled' % self.session.client_id)

    async def handle_suback(self, suback: SubackPacket):
        self.logger.debug('%s SUBACK unhandled' % self.session.client_id)

    async def handle_unsuback(self, unsuback: UnsubackPacket):
        self.logger.debug('%s UNSUBACK unhandled' % self.session.client_id)

    async def handle_pingresp(self, pingresp: PingRespPacket):
        self.logger.debug('%s PINGRESP unhandled' % self.session.client_id)

    async def handle_pingreq(self, pingreq: PingReqPacket):
        self.logger.debug('%s PINGREQ unhandled' % self.session.client_id)

    async def _handle_disconnet(self, disconnect: DisconnectPacket):
        self.logger.debug('%s DISCONNECT unhandled' % self.session.client_id)

    async def handle_disconnect(self, disconnect: DisconnectPacket):
        if not self._disconnecting:
            self._disconnecting = True
            await self._tg.spawn(self._handle_disconnect, disconnect)
        else:
            self.logger.debug('%s DISCONNECT ignored' % self.session.client_id)
        raise StopAsyncIteration  # end of the line

    async def handle_connection_closed(self):
        self.logger.debug('%s Connection closed unhandled' % self.session.client_id)

    async def handle_puback(self, puback: PubackPacket):
        packet_id = puback.variable_header.packet_id
        try:
            waiter = self._puback_waiters[packet_id]
            await waiter.set(puback)
        except KeyError:
            self.logger.warning("Received PUBACK for unknown pending message Id: '%d'", packet_id)
        except InvalidStateError:
            self.logger.warning("PUBACK waiter with Id '%d' already done", packet_id)

    async def handle_pubrec(self, pubrec: PubrecPacket):
        packet_id = pubrec.packet_id
        try:
            waiter = self._pubrec_waiters[packet_id]
            await waiter.set(pubrec)
        except KeyError:
            self.logger.warning("Received PUBREC for unknown pending message with Id: %d", packet_id)
        except InvalidStateError:
            self.logger.warning("PUBREC waiter with Id '%d' already done", packet_id)

    async def handle_pubcomp(self, pubcomp: PubcompPacket):
        packet_id = pubcomp.packet_id
        try:
            waiter = self._pubcomp_waiters[packet_id]
            await waiter.set(pubcomp)
        except KeyError:
            self.logger.warning("Received PUBCOMP for unknown pending message with Id: %d", packet_id)
        except InvalidStateError:
            self.logger.warning("PUBCOMP waiter with Id '%d' already done", packet_id)

    async def handle_pubrel(self, pubrel: PubrelPacket):
        packet_id = pubrel.packet_id
        try:
            waiter = self._pubrel_waiters[packet_id]
            await waiter.set(pubrel)
        except KeyError:
            self.logger.warning("Received PUBREL for unknown pending message with Id: %d", packet_id)
        except InvalidStateError:
            self.logger.warning("PUBREL waiter with Id '%d' already done", packet_id)

    async def handle_publish(self, publish_packet: PublishPacket):
        try:
            packet_id = publish_packet.variable_header.packet_id
            qos = publish_packet.qos

            incoming_message = IncomingApplicationMessage(packet_id, publish_packet.topic_name, qos, publish_packet.data, publish_packet.retain_flag)
            incoming_message.publish_packet = publish_packet
            if incoming_message.qos == QOS_0:
                await self._handle_message_flow(incoming_message)
            else:
                await self._reader_task.spawn(self._handle_message_flow, incoming_message)

            if self.session is not None:
                self.logger.debug("Message queue size: %d", self.session._delivered_message_queue.qsize())
        except CancelledError:
            pass

    async def wait_disconnect(self):
        if self._disconnect_waiter is not None:
            return await self._disconnect_waiter.wait()



# Copyright (c) 2015 Nicolas JOUANIN
#
# See the file license.txt for copying permission.


class DistMQTTException(Exception):
    """
    DistMQTT base exception
    """
    pass


class MQTTException(Exception):
    """
    Base class for all errors refering to MQTT specifications
    """
    pass


class CodecException(Exception):
    """
    Exceptions thrown by packet encode/decode functions
    """
    pass


class NoDataException(Exception):
    """
    Exceptions thrown by packet encode/decode functions
    """
    pass


class InvalidStateError(Exception):
    """
    Exception thrown by Future.
    """
    pass
